import datetime
import logging
import os.path
import re
import time
import urllib

from unicon.eal.dialogs import Dialog

from kick.device2.general.actions.power_bar import power_cycle_all_ports
from .patterns import FmcPatterns
from .statemachine import FmcStatemachine
from ...general.actions.basic import BasicDevice, BasicLine

try:
    from kick.kick_constants import KickConsts
except ImportError:
    from kick.miscellaneous.credentials import KickConsts

try:
    import kick.graphite.graphite as graphite
except ImportError:
    import kick.metrics.metrics as graphite

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 10
MAX_RETRY_COUNT = 3


class Fmc(BasicDevice):
    def __init__(self, hostname='firepower',
                 login_username='admin',
                 login_password=KickConsts.DEFAULT_PASSWORD,
                 sudo_password=KickConsts.DEFAULT_PASSWORD,
                 *args,
                 **kwargs):
        """Constructor for FMC.

        :param hostname: fmc hostname or fqdn e.g. FS2000-01 or FS2000-01.domain.com
        :param login_username: user name for login
        :param login_password: password for login
        :param fmc_root_password: root password for FMC
        :return: None

        Add *args and **kwargs so that the initiator can be invoked with
        additional (unused) arguments.

        """
        graphite.publish_kick_metric('device.fmc.init', 1)
        self.patterns = FmcPatterns(
            hostname=hostname,
            login_username=login_username,
            login_password=login_password,
            sudo_password=sudo_password)
        self.sm = FmcStatemachine(self.patterns)
        self.line_class = FmcLine

        self.set_default_timeout(DEFAULT_TIMEOUT)

    def log_checks(self, fmc_line, list_files=['/var/log/boot_*'],
                   search_strings=['fatal', 'error'], exclude_strings=[], timeout=300):
        """Wrapper function to Get logs for FMC.

        :param fmc_line: Instance of fmc line used to connect to FMC
        :param list_files: List of file paths for files to search in
        :param search_strings: List of keywords to be searched in the logs
        :param exclude_strings: List of keywords to be excluded in the logs
        :param timeout: The number for seconds to wait for log retrieval
        
        e.g list_files = ['/var/log/boot_1341232','/var/log/boot_*']
            search_strings = ['fatal','error', 'crash']
            exclude_strings = ['ssl_flow_errors', 'firstboot.S09']

        """

        self.sm.go_to('sudo_state', fmc_line.spawn_id, timeout=30)

        grep_command_list = []
        exclude_line = ''

        if exclude_strings:
            exclude_cmd = ['| grep -v {}'.format(string) for string in exclude_strings]
            exclude_line = ''.join(exclude_cmd)

        if list_files and search_strings:
            for file in list_files:
                for string in search_strings:
                    grep_command = "grep -Ii {} {} | sort -u {}".format(string, file, exclude_line)
                    grep_command_list.append(grep_command)
        try:
            output_log = fmc_line.execute_lines(("\n".join(grep_command_list)), timeout=timeout)
        except:
            output_log = "Log retrieval command timed out"

        logger.info("""
            ***********************************************************

            Logs for the requested files in FMC are : -

            {}

            ***********************************************************
            """.format(output_log))

    def define_contexts(self, contexts=None):
        raise NotImplementedError

    def ssh_vty(self, ip, port, username='admin', password=KickConsts.DEFAULT_PASSWORD,
                timeout=None, line_type='ssh_vty', rsa_key=None):
        return super().ssh_vty(ip, port, username=username, password=password,
                               timeout=timeout, line_type=line_type, rsa_key=rsa_key)


class FmcLine(BasicLine):
    def __init__(self, spawn, sm, type, timeout=None):
        """ Constructor of FmcLine

        :param spawn: spawn_id of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh', 'telnet'
        :return: object of the line

        """

        try:
            super().__init__(spawn, sm, type, timeout)
        except:
            if self.type in ['ssh_vty', 'ssh_cimc']:
                raise
            # keep the FmcLine 'alive' in case the power cycle function is used
            # in order to start a baseline
            logger.info("=== Device was accessed via console port")
            logger.info("=== Failed to go to a known state. Though we keep the FmcLine "
                        "'alive' in case the power cycle function is used")

    def expert_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        """Run a command as 'admin' and return its output

        :param cmd: command to be executed as 'admin' given as a string
        :param timeout: in seconds; how long to wait for command output
        :param exception_on_bad_command: True/False - whether to raise an exception
                on a bad command
        :return: output as a string

        """

        self.go_to('admin_state')
        return self.execute(cmd, timeout, exception_on_bad_command)

    def changeto_context(self, ctx, timeout=60):
        """ Not implemented for FmcLine class"""

        raise NotImplementedError

    def find_matched_str(self,
                         expected_str,
                         return_data,
                         matched_whole_line=False):
        """Find matched line or substring from return_data.

        :param expected_str: expected string for the output of cmd executed
        :param return_data: command output
        :param matched_whole_line: True if want to match the whole line, otherwise, match the substring
        :return: matched_obj

        """

        matched_obj = False
        if matched_whole_line:
            for line in return_data.split('\n'):
                line = line.strip()
                if line == '':
                    continue
                logger.debug('line=' + line)
                if line == expected_str:
                    matched_obj = line
                    break
        else:
            matched_obj = re.search(expected_str, return_data)
        return matched_obj

    def execute_and_verify(self, cmd,
                           expected_str,
                           cmd_set_config='',
                           timeout=60,
                           interval=2,
                           timeout_total=60,
                           retry_count=MAX_RETRY_COUNT,
                           matched_whole_line=False):
        """Execute a command, verify the output.

        set the config if the output is not expected,
        The command will be run and retried for no more than retry_count,
        and no more than timeout_total in seconds
        Assert if expected_str is not found in the output of cmd

        :param cmd: command to be executed
        :param expected_str: expected string for the output of cmd executed
        :param cmd_set_config: command to be executed if expected_str is not seen, if cmd_set_config is an empty string,
                                the function is waiting only
        :param timeout: timeout for execute() in seconds
        :param interval: wait time for each retry in seconds
        :param timeout_total: total wait time in seconds
        :param retry_count: Max count for retries
        :param matched_whole_line: True if want to match the whole line
        :return: None

        """

        logger.info("""Command: {}
            Expected String: {}
            Command to set configuration: {}
        """.format(cmd, expected_str, cmd_set_config))
        count = 0
        expected = False
        start_time = datetime.datetime.now()
        elapsed_time = 0
        while count < retry_count and elapsed_time < timeout_total:
            return_data = self.execute(cmd, timeout=timeout)
            matched_obj = self.find_matched_str(expected_str, return_data,
                                                matched_whole_line)
            if matched_obj is None:
                if cmd_set_config is not '':
                    # Execute the command to get expected_str
                    self.execute(cmd_set_config, timeout=timeout)
                    self.execute('commit', timeout=timeout)
                count += 1
                now = datetime.datetime.now()
                elapsed_time = (now - start_time).total_seconds()
                time.sleep(interval)
            else:
                expected = True
                break
        assert expected, "Command: {}, Expected String: {} " \
                         "not found".format(cmd, expected_str)

    def validate_mysql_process(self):
        """ Validate mysql process"""

        self.go_to('admin_state')
        self.execute_and_verify(cmd='ps -ef | grep mysql | grep -v grep',
                                expected_str='mysqld',
                                timeout=120,
                                interval=5,
                                timeout_total=300,
                                retry_count=60)
        self.execute_and_verify(cmd='ls /var/run/mysql/',
                                expected_str='mysql\.sock',
                                timeout=120,
                                interval=5,
                                timeout_total=300,
                                retry_count=60)
        self.execute_and_verify(cmd='ls /var/run/mysql/',
                                expected_str='mysqld\.pid',
                                timeout=120,
                                interval=5,
                                timeout_total=300,
                                retry_count=60)

    def configure_network(self, mgmt_gateway, mgmt_ip, mgmt_netmask,
                          mgmt_ip6=None, mgmt_prefix=None, mgmt_gateway6=None,
                          change_password=True):
        """
        Configure the network setting of the management interface

        :param mgmt_ip: management interface ip address
        :param mgmt_netmask: management interface netmask
        :param mgmt_gateway: management interface gateway
        :param mgmt_ip6: management interface IPv6 Address
        :param mgmt_prefix: management interface IPv6 Prefix
        :param mgmt_gateway6: management interface IPv6 Gateway
        :param change_password: Flag to change the password after baseline
        :return:
        """
        #Configure network
        try:
            self.spawn_id.sendline('expert')
        except:
            pass
        self.spawn_id.sendline('sudo su -')
        self.spawn_id.expect("Password: ", 10)
        self.spawn_id.sendline(self.sm.patterns.default_password)
        self.go_to('any')
        self.spawn_id.sendline('configure-network')
        d5 = Dialog([
            ['Do you wish to configure IPv4', 'sendline({})'.format('y'),
             None, True, False],
            ['Management IP address', 'sendline({})'.format(mgmt_ip),
             None, True, False],
            ['Management netmask', 'sendline({})'.format(mgmt_netmask),
             None, True, False],
            ['Management default gateway', 'sendline({})'.format(mgmt_gateway),
             None, False, False],
        ])
        d5.process(self.spawn_id, timeout=180)
        d6 = Dialog([
            ['Are these settings correct', 'sendline({})'.format('y'),
             None, False, False],
        ])
        d6.process(self.spawn_id, timeout=180)
        if mgmt_ip6 is None:
            d7 = Dialog([
                ['Do you wish to configure IPv6', 'sendline({})'.format('n'),
                 None, True, False],
                [self.sm.get_state('sudo_state').pattern, None, None, False, False],
            ])
            d7.process(self.spawn_id, timeout=300)
        else:
            d7 = Dialog([
                ['Do you wish to configure IPv6', 'sendline({})'.format('y'),
                 None, True, False],
                ['Do you wish to router autoconfiguration', 'sendline({})'.format('n'),
                 None, True, False],
                ['Management IP address', 'sendline({})'.format(mgmt_ip6),
                 None, True, False],
                ['Management prefix', 'sendline({})'.format(mgmt_prefix),
                 None, True, False],
                ['Management default gateway', 'sendline({})'.format(mgmt_gateway6),
                 None, False, False],
            ])
            d7.process(self.spawn_id, timeout=180)
            d8 = Dialog([
                ['Are these settings correct', 'sendline({})'.format('y'),
                 None, True, False],
                [self.sm.get_state('sudo_state').pattern, None, None, False, False],
            ])
            d8.process(self.spawn_id, timeout=300)

        self.execute_and_verify(cmd='ifconfig',
                                expected_str=mgmt_ip)
        if mgmt_ip6 is not None:
            self.execute_and_verify(cmd='ifconfig',
                                    expected_str=mgmt_ip6)

        if change_password:
            logger.info('=== Changing default password')
            self.spawn_id.sendline('passwd {}'.format(self.sm.patterns.username))
            d9 = Dialog([
                ['New UNIX password', 'sendline({})'.format(self.sm.patterns.login_password),
                 None, True, False],
                ['Retype new UNIX password', 'sendline({})'.format(self.sm.patterns.login_password),
                 None, True, False],
                ['passwd: password updated successfully', None, None, False, False],
            ])
            try:
                d9.process(self.spawn_id, timeout=120)
            except TimeoutError:
                logger.error('Changing password for {} user has failed'.format(self.sm.patterns.username))
                raise Exception('Error while changing default password for {} user'.format(self.sm.patterns.username))

    def validate_version(self, iso_file_name,private_iso=False):
        """Checks if the installed version matches the version from iso filename.
        Make sure you are in 'sudo_state' before calling this method.

        :param iso_file_name: iso filename under http_link,
               e.g. 'Sourcefire_Defense_Center_M4-6.2.0-362-Autotest.iso'
        :param private_iso: Flag to validate version for private iso
        :return: None

        """

        self.go_to('sudo_state')
        version = self.execute('cat /etc/sf/ims.conf | egrep "SWVERSION"', 30)
        build = self.execute('cat /etc/sf/ims.conf | egrep "SWBUILD"', 30)
        # extract version and build from iso file name
        if private_iso:
            expected_version = re.search('[\d.]+', iso_file_name.split('-')[-2]).group(0)
            expected_build = re.search('\d+', iso_file_name.split('-')[-1]).group(0)
        else:
            expected_version = re.search('[\d.]+', iso_file_name.split('-')[-3]).group(0)
            expected_build = re.search('\d+', iso_file_name.split('-')[-2]).group(0)

        if version.split('=')[1] == expected_version and build.split('=')[1] == expected_build:
            logger.info('>>>>>> Version {}, Build {} match {} {}'.format(version.split('=')[1],
                                                                         build.split('=')[1],
                                                                         expected_version,
                                                                         expected_build))
            logger.info('Installed FMC version validated')
            return True
        else:
            logger.error('Exception: not the same version and build')
            raise RuntimeError(
                '>>>>>> Version {}, Build {} does not match {} {}'.format(version.split('=')[1],
                                                                          build.split('=')[1],
                                                                          expected_version,
                                                                          expected_build))

    def power_cycle(self, power_bar_server, power_bar_port, power_bar_user='admn', power_bar_pwd='admn'):
        """Power cycle the device using power_bar

        :param power_bar_server: comma-separated string of IP addresses of the PDU's
        :param power_bar_port: comma-separated string of power ports on the PDU's
        :param power_bar_user: comma-separated string of usernames for power bar servers
        :param power_bar_pwd:  comma-separated string of passwords for power bar servers
        :return:
        """
        if power_bar_server == '' or power_bar_port == '':
            raise AssertionError("Cannot power cycle. powerbar server or powerbar port info missing")

        power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd)

    def _move_from_lilo_boot_menu_to_lilo_os(self):
        """Go to LILO boot menu to lilo os """

        for i in range(0, 30):
            # Keep sending control+c till device moved to lilo os state
            self.spawn_id.send("\x03")
            try:
                self.spawn_id.expect("Please press Enter to activate this console", timeout=10)
                self.spawn_id.send("\r")
                self.spawn_id.expect(self.sm.patterns.prompt.lilo_os_prompt, timeout=10)
                break
            except:
                pass

        if i >= 30:
            msg = "Could not go to LILO OS prompt from LILO boot prompt"
            logger.error(msg)
            raise RuntimeError(msg)

    def _try_to_goto_prompt(self, prompt, num_of_tries=10, raise_exception=True):
        """ Try to go to given prompt in given number of attempts

        :param prompt: Prompt
        :param num_of_tries: Number of attempts to go to given prompt
        :param raise_exception: Raise exception on failure

        :return: return 0 - success
                        1 - failure

        """

        for i in range(1, num_of_tries+1):
            try:
                self.go_to(prompt)
                break
            except:
                pass

        if i >= num_of_tries:
            msg = "Could not go to %s even after %s tries" % (prompt, num_of_tries)
            logger.error(msg)
            if raise_exception:
                raise RuntimeError(msg)

            return 1

        return 0

    def _install_from_iso(self, iso_url, uut_ip, uut_netmask, uut_gateway, mgmt_intf="eth0",
                          usb_loc="/dev/sdb1", timeout=7200):
        """ Install image from given ISO

        :param iso_url: ISO URL
        :param uut_ip:  Device IP
        :param uut_netmask: Device netmask
        :param uut_gateway: Devcie Gateway
        :param mgmt_intf: Management Interface name (default: eth0)
        :param usb_loc: USB location in devcie (default: /dev/sdb1)
        :return: None

        """

        lilo_os_prompt = self.sm.patterns.prompt.lilo_os_prompt

        self.spawn_id.sendline("")
        self.spawn_id.expect(self.sm.patterns.prompt.lilo_os_prompt)

        self.execute("umount /mnt/usb")
        output = self.execute("mount %s /mnt/usb" % usb_loc)
        if '%s is already mounted' % usb_loc in output:
            logger.error("Failed mounting {} on /mnt/usb".format(usb_loc))
            raise RuntimeError("Failed mounting {} on /mnt/usb".format(usb_loc))
        self.execute("/sbin/ifconfig %s %s netmask %s" % (mgmt_intf, uut_ip, uut_netmask))
        self.execute("route add default gw %s" % uut_gateway)

        # Get ISO name, path and server ip from URL path
        url_data = urllib.parse.urlparse(iso_url)
        install_srv = url_data.netloc
        iso_name = os.path.basename(url_data.path)
        iso_path = os.path.dirname(url_data.path)[1:]
        
        self.execute("mkdir /mnt/usb/configs")
        self.execute("""echo "SRV=%s
        dirpath=%s
        TRANS=httpsvr
        choicedevice=%s
        IPCONF=dhcp
        IPFAM=ipv4
        choiceiso=%s
        " > /mnt/usb/configs/default_config.conf""" % (install_srv, iso_path, mgmt_intf, iso_name))

        config_output = self.execute("cat /mnt/usb/configs/default_config.conf")
        if 'No such file or directory' in config_output:
            logger.error('Failed creating default_config file')
            raise RuntimeError('Failed creating default_config file')

        self.execute("umount /cdrom  2> /dev/null")
        self.execute("rm -f /mnt/tmpfs/*")
        self.execute("umount /mnt/tmpfs 2> /dev/null")
        self.execute("umount /mnt/hd/ 2> /dev/null")
        self.execute("umount /mnt/usb 2> /dev/null")

        self.execute("/etc/rc.d/install.sh --partition-only", timeout=240)

        # Download image into /mnt/var and initiate installation
        self.execute("cd /mnt/var/")

        for i in range(1, 11):
            try:
                d1 = Dialog([
                    ['The file is already fully retrieved; nothing to do.', None, None, False, False],
                    ["'%s' saved" % iso_name, None, None, False, False],
                ])
                self.spawn_id.sendline("wget --no-check-certificate -c %s" % iso_url)
                d1.process(self.spawn_id, timeout=300)
                break
            except:
                pass

        if i >= 10:
            msg = "Failed to down load %s in 10 attempts with timeout 5 mins" % "wget -c %s" % iso_url
            logger.error(msg)
            raise RuntimeError(msg)

        self.execute("mount -t iso9660 -o loop /mnt/var/%s /cdrom" % iso_name, timeout=90)

        # Install steps
        self.execute("umount /mnt/usb &> /dev/null")

        self.spawn_id.sendline("/cdrom/install/menu.sh")

        # wait for menu.sh script to be launched
        d = Dialog([
            ['This might take some time', None, None, False, False],
            # needed for older builds (6.1.0)
            ['All data will be destroyed', None, None, False, False],
            # needed for older builds (6.0.0)
            ['The USB device was successfully imaged', None, None, False, False],
        ])
        try:
            d.process(self.spawn_id, timeout=60)
        except Exception as e:
            logger.info("menu script has not started after 1 min. It failed with {}".format(e))
            raise Exception("menu.sh script has not started after 1 min")

        d0 = Dialog([
            ["The system will restart after you press enter.", 'sendline()', None, True, False],
            ['Restore the system.*yes/no.*:', 'sendline(yes)', None, True, False],
            ['Delete license and network settings.*yes/no.*:', 'sendline(no)', None, True, False],
            ['Are you sure.*yes/no.*:', 'sendline(yes)', None, True, False],
            [lilo_os_prompt, None, None, False, False],
            [self.sm.patterns.prompt.lilo_boot_menu_prompt, None, None, False, False],
        ])
        res = d0.process(self.spawn_id, timeout=600)

        if self.sm.patterns.prompt.lilo_boot_menu_prompt in res.match_output:
            logger.info("Device moved back to lilo boot menu")
            return 1

        if "root@(none)" in res.match_output:
            self.execute("umount /cdrom")
            self.execute("losetup -d /dev/loop0")
            self.execute("rm -rf /mnt/var/*.iso")

            # Reboot system to complete install process
            self.spawn_id.sendline("reboot")
            d0 = Dialog([
                [self.sm.patterns.prompt.lilo_boot_menu_prompt, None, None, False, False],
                # needed for M3 to redirect the output to the console
                ['\nboot:', 'sendline({})'.format('1'), None, True, False],
                ['.*login:', 'sendline({})'.format(self.sm.patterns.username), None, True, False],
                ['Password:', 'sendline({})'.format(self.sm.patterns.default_password), None, False, False]
            ])
            res = d0.process(self.spawn_id, timeout=timeout)

            if self.sm.patterns.prompt.lilo_boot_menu_prompt in res.match_output:
                logger.info("Device moved back to lilo boot menu")
                return 1
            elif 'Password:' in res.match_output:
                logger.info("Install is done and got firepower login: prompt")
                self._first_login()
                return 0

        logger.error("Unknown state. It should never be here")

        return 2

    def replace_asa_image(self, source_location, pwd, timeout=300):
        """ Not implemented for FmcLine class"""

        raise NotImplementedError

    def _first_login(self):

        d0 = Dialog([
            ['\(current\) UNIX password', 'sendline({})'.format(self.sm.patterns.default_password), None, True, False],
            ['New UNIX password', 'sendline({})'.format(self.sm.patterns.login_password), None, True, False],
            ['Retype new UNIX password', 'sendline({})'.format(self.sm.patterns.login_password), None, False, False],
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
            ['--More--', 'send(\x20)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ", 'sendline()', None, False, False],
                ])

        accept_eula_dialog = Dialog([
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
            ['--More--', 'send(\x20)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ", 'sendline()', None, False, False],
            ['Enter new password', 'sendline({})'.format(self.sm.patterns.login_password), None, True, False],
            ['Confirm new password', 'sendline({})'.format(self.sm.patterns.login_password), None, False, False],
        ])
        try:
            d0.process(self.spawn_id, timeout=90)
            try:
                accept_eula_dialog.process(self.spawn_id, timeout=90)
            except TimeoutError:
                logger.info("=== EULA was not displayed")
            logger.info("=== Successfully completed the first login for fmc")
        except TimeoutError:
            logger.warning('Changing password for {} user was not required at this step'.
                           format(self.sm.patterns.login_username))
            self.spawn_id.sendline()

    def configuration_wizard(self, ipv4_mode, ipv6_mode, ipv4, ipv4_netmask, ipv4_gateway,
                             ipv6, ipv6_prefix, ipv6_gateway, dns_servers, search_domains,
                             ntp_servers):
        """ Handle FMC configuration wizard

        :param ipv4_mode: IPv4 configuration mode - 'static' ('manual') or 'dhcp'
        :param ipv4: IPv4 address of the management interface
        :param ipv4_netmask: IPv4 netmask of the management interface
        :param ipv4_gateway: IPv4 gateway of the management interface
        :param ipv6: IPv6 address of the management interface
        :param ipv6_prefix: IPv6 prefix of the management interface
        :param ipv6_gateway: IPv6 gateway of the management interface
        :param ipv6_mode: IPv6 configuration mode - 'dhcp', 'router' or 'manual'
        :param dns_servers: a comma-separated string of DNS servers
        :param search_domains: a comma-separated string of search domains
        :param ntp_servers: a comma-separated string of NTP servers
        :return: True if network configuration was done, False otherwise
        """
        if ipv4_mode in ["static", "manual"]:
            ipv4_mode_new = "manual"
            ipv4_mode_old = "static"
        else:
            ipv4_mode_old = "dhcp"
            ipv4_mode_new = "dhcp"

        config_dialog = Dialog([
            [self.sm.patterns.prompt.fireos_prompt, 'sendline()', None, False, False],
            [self.sm.patterns.prompt.admin_prompt, 'sendline()', None, False, False],
            [self.sm.patterns.prompt.sudo_prompt, 'sendline()', None, False, False],
            ['Enter a hostname or fully qualified domain name for this system ',
             'sendline({})'.format(self.sm.patterns.fqdn), None, True, False],
            ['Enter a fully qualified hostname for this system ',
             'sendline({})'.format(self.sm.patterns.fqdn), None, True, False],
            ['Configure IPv4 via DHCP or manually', 'sendline({})'.format(ipv4_mode_new),
             None, True, False],
            ['Configure IPv4 via DHCP or static configuration', 'sendline({})'.format(ipv4_mode_old),
             None, True, False],
        ])

        if ipv4_mode is 'static':
            config_dialog.append(['Enter an IPv4 address for the management interface',
                                  'sendline({})'.format(ipv4), None, True, False])
            config_dialog.append(['Enter an IPv4 netmask for the management interface',
                                  'sendline({})'.format(ipv4_netmask), None, True, False])
            config_dialog.append(['Enter the IPv4 default gateway for the management interface',
                                  'sendline({})'.format(ipv4_gateway), None, True, False])

        if ipv6_mode:
            config_dialog.append(['Do you want to configure IPv6', 'sendline(y)', None, True, True])
            config_dialog.append(['Configure IPv6 via DHCP, router, or manually', 'sendline({})'.format(ipv6_mode), None, True, True])
        else:
            config_dialog.append(['Do you want to configure IPv6', 'sendline(n)', None, True, True])

        if ipv6_mode is 'manual':
            config_dialog.append(['Enter the IPv6 address for the management interface',
                                  'sendline({})'.format(ipv6), None, True, False])
            config_dialog.append(['Enter the IPv6 address prefix for the management interface',
                                  'sendline({})'.format(ipv6_prefix), None, True, False])
            config_dialog.append(['Enter the IPv6 gateway for the management interface',
                                  'sendline({})'.format(ipv6_gateway), None, True, False])

        config_dialog.append(['Enter a comma-separated list of DNS servers', 'sendline({})'.format(dns_servers),
                              None, True, False])
        config_dialog.append(['Enter a comma-separated list of search domains', 'sendline({})'.format(search_domains),
                              None, True, False])
        config_dialog.append(['Enter a comma-separated list of NTP servers', 'sendline({})'.format(ntp_servers),
                              None, True, False])
        config_dialog.append(['Are these settings correct', 'sendline(y)', None, True, False])
        config_dialog.append(['Updated network configuration', None, None, True, False])

        response = config_dialog.process(self.spawn_id, timeout=900)

        self.spawn_id.sendline()
        self.go_to('any')

        if 'Updated network configuration' in response.match_output:
            logger.info("Network configuration completed")
            return True

        return False



from .asa import Asa, AsaLine
from .asa_config import AsaConfig
from .asa_cluster import AsaCluster
from .asa_gtp import AsaGtpConfig
from .asa_diameter import AsaDiameterConfig
from .asa_ha import AsaHAActiveStandby, AsaHAActiveActive



from ...kp.actions.statemachine import KpStateMachine
from unicon.statemachine import Path


class WmStateMachine(KpStateMachine):

    def __init__(self, patterns):
        """Initializer of SspStateMachine."""
        super().__init__(patterns)

    def create(self):

        super().create()

        # fireos to fxos transition has changed on child Wm class
        new_path_ftd_to_fxos = Path(self.get_state('fireos_state'), self.get_state('fxos_state'), "exit")

        # Add path to the State Machine
        self.remove_path(self.get_state('fireos_state'), self.get_state('fxos_state'))
        self.add_path(new_path_ftd_to_fxos)



"""This is nothing but a place holder, to save instance-specific variables.

For this to work, the __init__() of the class has to be written as below:
 1. Initialize KpConstants.hostname, login_username and login_password
 2. Create the state machine.

In step 2 above, the state machine is written with the assumption that
KpConstants is properly initialized. That's why the order in __init__()
above can not be reversed.

"""

class KpConstants:
    pass



from unicon.eal.dialogs import Dialog
from kick.device2.ftd5500x.actions.dialogs import Ftd5500xDialog


class KpDialogs:
    # Dialogs for state transitions
    def __init__(self, patterns):
        """Initializer of KpDialogs."""
        self.patterns = patterns
        self.d_expert_to_sudo = Dialog([[self.patterns.prompt.password_prompt,
                                         'sendline({})'.format(self.patterns.sudo_password),
                                         None, True, True], ])
        self.d_ftd_to_fxos = Dialog([
            ["Please enter 'exit' to go back", 'sendline(exit)', None, True, False], ])

        self.ftd_dialogs = Ftd5500xDialog(patterns)



from unicon.eal.dialogs import Dialog
from .constants import WebserverConstants


class WebserverDialog:
    ssh_connect_dialog = Dialog([
        ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
        ['Password:', 'sendline_ctx(password)', None, True, False],
        ['Last login:', None , None, True, False],
    ])



import logging
import os
import re
import time
import sys
import traceback

try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric
from ..actions.statemachine import WmStateMachine
from ...kp.actions import Kp, KpLine
from ...kp.actions.patterns import KpPatterns

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

logger = logging.getLogger(__name__)
MAX_RETRY_COUNT = 3
DEFAULT_TIMEOUT = 60


class Wm(Kp):
    def __init__(self, hostname, login_username='admin',
                 login_password='Admin123', sudo_password='Admin123',
                 *args,
                 **kwargs):
        """
		Wm instance constructor
		:param hostname: hostname of the FXOS
        :param login_username: user name for login
        :param login_password: password for login
        :param sudo_password: root password for FTD
        :kwargs
        :param: 'config_hostname': output of 'show running-config hostname'

		"""

        super(Kp, self).__init__()
        publish_kick_metric('device.wm.init', 1)
        self.set_default_timeout(DEFAULT_TIMEOUT)

        # set hostname, login_username and login_password
        config_hostname = kwargs.get('config_hostname', 'firepower')
        self.patterns = KpPatterns(hostname, login_username, login_password, sudo_password, config_hostname)

        # create the state machine that contains the proper attributes.
        self.sm = WmStateMachine(self.patterns)

        # important: set self.line_class so that a proper line can be created
        # by ssh_console(), etc.
        self.line_class = WmLine
        logger.info("Done: Wm instance created")


class WmLine(KpLine):
    def __init__(self, spawn_id, sm, type, chassis_line=True, timeout=None):
        """Constructor of WmLine instance

		:param spawn_id: spwan connection
		:param sm: state machine instance
		:param type: 'telnet' or 'ssh'
		:param chassis_line: True if ssh console connection is used
                             False if ssh to FTD directly
		:param timeout: in seconds
		"""

        super(KpLine, self).__init__(spawn_id, sm, type, timeout=timeout)
        self.chassis_line = chassis_line
        self.line_type = 'WmLine'
        self.change_password_flag = False
        self.power_cycle_flag = False

        if self.chassis_line and self.sm.current_state is not 'rommon_state':
            ##### There were defects CSCvq86739, CSCvq96757 and CSCvq93377 where device prompt and command were lagging
            #### This sleep provide to wait for prompt to come
            time.sleep(2)
            self.go_to('any')
            self.init_terminal(determine_state=False)

    def baseline(self, tftp_server, rommon_file,
                 uut_hostname, uut_username, uut_password,
                 uut_ip, uut_netmask, uut_gateway,
                 dns_servers, search_domains,
                 fxos_url, ftd_version, file_server_password="",
                 power_cycle_flag=False,
                 mode='local',
                 uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                 manager=None, manager_key=None, manager_nat_id=None,
                 firewall_mode='routed', timeout=3600, reboot_timeout=300, only_ftd=False,dhcp=False):

        """ Method used to baseline the Westminster device.
        Does the following steps:
        - assigns an ip address from mgmt network to the sensor in rommon mode
        - tftp and boots an fxos image
        - from within FXOS downloads and installs a FTD image

        :param tftp_server: tftp server to get rommon and fxos images
        :param rommon_file: rommon build,
               e.g. 'asa/cache/Development/6.4.0-10138/installers/fxos-k8-fp2k-lfbff.82.5.1.893i.SSB'
        :param uut_hostname: hostname of the FTD
        :param uut_username: User Name to login to FTD
        :param uut_password: Password to login to FTD
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_servers: DNS Servers
        :param search_domains: Search Domains
        :param fxos_url: FXOS+FTD image url,
               e.g. 'tftp://10.89.23.30/asa/cache/Development/6.4.0-10138/installers/
               cisco-ftd-fp1k.6.4.0-10138.SSB'
        :param ftd_version: ftd version, e.g. '6.4.0-10138'
        :param file_server_password: if use scp protocol, this is the password to
               download the image
        :param power_cycle_flag: if True power cycle the device before baseline
        :param mode: the manager mode ('local', 'remote')
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param firewall_mode: the firewall mode ('routed', 'transparent', 'ngips')
        :param timeout: in seconds; time to wait for installing the security package;
                        default value is 3600s
        :param reboot_timeout: in seconds; time to wait for system to restart;
                        default value is 300s
        :param only_ftd: Flag to install only FTD , Default is "False"
        :param dhcp: Flag to take the network configuration from DHCP , Default is "False"
        :return: dhcp_ip

        """
        publish_kick_metric('device.wm.baseline', 1)
        ######Install FTD package alone if flag is set ####
        if only_ftd:
            dhcp_ip = self.baseline_ftd(uut_hostname=uut_hostname,
                              uut_password=uut_password,
                              uut_ip=uut_ip, uut_netmask=uut_netmask, uut_gateway=uut_gateway,
                              dns_servers=dns_servers, search_domains=search_domains,
                              fxos_url=fxos_url, ftd_version=ftd_version,
                              file_server_password=file_server_password,
                              mode=mode,
                              uut_ip6=uut_ip6, uut_prefix=uut_prefix, uut_gateway6=uut_gateway6,
                              firewall_mode=firewall_mode, timeout=timeout,dhcp=dhcp)
        ######Install fxos/ftd installation with format everything on device ####
        else:
            dhcp_ip = self.baseline_fp2k_ftd(tftp_server=tftp_server,
                                   rommon_file=rommon_file, uut_hostname=uut_hostname,
                                   uut_username=uut_username, uut_password=uut_password,
                                   uut_ip=uut_ip, uut_netmask=uut_netmask, uut_gateway=uut_gateway,
                                   dns_servers=dns_servers, search_domains=search_domains,
                                   fxos_url=fxos_url, ftd_version=ftd_version,
                                   file_server_password=file_server_password,
                                   power_cycle_flag=power_cycle_flag, mode=mode,
                                   uut_ip6=uut_ip6, uut_prefix=uut_prefix, uut_gateway6=uut_gateway6,
                                   manager=manager, manager_key=manager_key,
                                   manager_nat_id=manager_nat_id, firewall_mode=firewall_mode, timeout=timeout,
                                   reboot_timeout=reboot_timeout,dhcp=dhcp)
        return dhcp_ip

    def baseline_by_branch_and_version(self, site, branch, version,
                                       uut_ip=None, uut_netmask=None, uut_gateway=None,
                                       dns_server='', serverIp='', tftpPrefix='', scpPrefix='', docs='', only_ftd=False,
                                       dhcp=False,
                                       **kwargs):
        """Baseline Westminster device by specifying branch and version.
        Looks for needed files on devit-engfs server, copies them to the local kick server
        and uses them to baseline the device.

        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature', 'Development'
        :param version: software build-version, e,g, 6.4.0-10138
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param \**kwargs:
            :Keyword Arguments, any of below optional parameters:
        :param uut_hostname: Device Host Name in the prompt
        :param uut_username: User Name to login to uut
        :param uut_password: Password to login to uut
        :param search_domains: search domain, defaulted to 'cisco.com'
        :param file_server_password: if use scp protocol, this is the password to
               download the image
        :param power_cycle_flag: if True power cycle the device before baseline
        :param mode: the manager mode (local, remote)
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param firewall_mode: the firewall mode (routed, transparent, ngips)
        :param timeout: in seconds; time to wait for installing the security package;
                        default value is 3600s
        :param reboot_timeout: in seconds; time to wait for system to restart;
                        default value is 300s
        :param only_ftd: Flag to installed ony FTD, default is "False"
        :param dhcp: Flag to baselie with DHCP client enabled, default is "False"
        :return: dhcp_ip

        """

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            scp_prefix = scpPrefix
            files = docs
        else:
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 'wm', branch, version)

        try:
            rommon_file = self._get_rommon_file(branch, version, files)
        except RuntimeError:
            rommon_file = [file for file in files
                           if file.startswith('fxos-k8-fp2k-lfbff') or file.startswith('fxos-k8-lfbff')][0]
        except Exception as e:
            raise Exception('Got {} while getting fxos file'.format(e))
        rommon_image = os.path.join(tftp_prefix, rommon_file)
        files.remove(rommon_file)
        pkg_file = files[0]
        ftd_file = os.path.join('tftp://{}'.format(server_ip), tftp_prefix, pkg_file)

        if not kwargs.get('ftd_version'):
            ftd_version = re.findall(r'[\d.]+-[\d]+', version)[0]
            kwargs['ftd_version'] = ftd_version

        kwargs['tftp_server'] = server_ip
        kwargs['rommon_file'] = rommon_image
        kwargs['fxos_url'] = ftd_file
        kwargs['uut_ip'] = uut_ip
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['dns_servers'] = dns_server
        kwargs['search_domains'] = kwargs.get('search_domains', 'cisco.com')
        kwargs['uut_hostname'] = kwargs.get('uut_hostname', 'firepower')
        kwargs['uut_username'] = kwargs.get('uut_username', 'admin')
        kwargs['uut_password'] = kwargs.get('uut_password', 'Admin123')
        kwargs['only_ftd'] = only_ftd
        kwargs['dhcp'] = dhcp
        dhcp_ip = self.baseline(**kwargs)
        return dhcp_ip

    def _get_rommon_file(self, branch, version, files):
        """
        Find the proper rommon_file.

        For old releases, it is always fxos-k8-fp2k-lfbff..., or fxos-k8-lfbff..., regardless
        of KP or WM. In new releases, it will be fxos-k8-fp1k for KP, fxos-k8-fp2k for WM.

        :param branch: such as Release
        :param version: such as 6.4.0-102
        :param files: list of files from prepare_installation_files()
                For example: ['cisco-ftd-fp1k.6.4.0-102.SPA',
                              'fxos-k8-fp1k-lfbff.2.6.1.133.SPA',
                              'fxos-k8-fp1k-lfbff.2.6.1.133i.SSB']
        :return:
        """

        if self._is_split_version(version):
            p = 'fxos-k8-fp1k-lfbff.*'
        else:
            p = 'fxos-k8-(fp2k-)?lfbff.*'

        if branch == 'Release':
            p += 'SPA'
        else:
            p += 'SSB'

        for f in files:
            if re.search(p, f):
                return f
        else:
            raise RuntimeError("Cannot file rommon_file in list of {}".format(files))



"""Copyright (c) 2017 Cisco Systems, Inc.
Name:
    asa_ha.py
Usage:
    Submodule for Legacy ASA HA.
Author:
    raywa
"""

import re
import time

from unicon.eal.dialogs import Dialog

from .constants import ASAHAStates, ASAHALinkNames

class AsaHAActiveStandby():
    """Class ASA active/standby HA
    """
    def __init__(self, asa_instance1, asa_instance2, init_ha=True):
        """Initializer of AsaHAActiveStandby.
        By default, instance1 will be primary and instance2 will be secondary.

        :param asa_instance1: connection instance1 that created by AsaConfig
        :param asa_instance2: connection instance2 that created by AsaConfig
        :return: None

        """
        self.asa_instance1 = asa_instance1
        self.asa_instance2 = asa_instance2
        self.primary = asa_instance1
        self.secondary = asa_instance2
        self.active = None
        self.standby = None

        if init_ha:
            # Configure primary unit
            self.config_as_primary(self.primary)

            folink_info = self.primary.topo.failover.failover_link
            if 'vlan' in folink_info and folink_info.vlan:
                folink_intf = '{hardware}.{vlan}'.format(hardware=folink_info.interface, vlan=folink_info.vlan)
                cmd = "interface %s\nno shutdown\n" % folink_info.interface
                cmd += "interface %s\nvlan %s" % (folink_intf, folink_info.vlan)
                self.primary.config(cmd, ctx = 'system')
                self.secondary.config(cmd, ctx='system')
            else:
                folink_intf = '{hardware}'.format(hardware=folink_info.interface)

            folink_ip = folink_info.ip_addr
            folink_mask = folink_info.netmask
            folink_standby = folink_info.ip_addr_standby
            folink_name = folink_info.get('link_name', ASAHALinkNames.DEFAULT_FAILOVER_LINK.value)
            self.config_folink(self.primary, folink_intf, folink_ip, folink_mask, folink_standby, folink_name)
            
            # Stateful link is optional.
            # Check existence in topo first
            if 'state_link' in self.primary.topo.failover:
                statelink_info = self.primary.topo.failover.state_link
                if 'vlan' in statelink_info and statelink_info.vlan:
                    statelink_intf = '{hardware}.{vlan}'.format(hardware=statelink_info.interface, vlan=statelink_info.vlan)
                    cmd = "interface %s\nno shutdown\n" % statelink_info.interface
                    cmd += "interface %s\nvlan %s" % (statelink_intf, statelink_info.vlan)
                    self.primary.config(cmd, ctx='system')
                    self.secondary.config(cmd, ctx='system')
                else:
                    statelink_intf = '{hardware}'.format(hardware=statelink_info.interface)

                if statelink_intf != folink_intf:
                    state_link_name = statelink_info.get(
                        'link_name', ASAHALinkNames.DEFAULT_STATE_LINK.value)
                    statelink_ip = statelink_info.get('ip_addr', None)
                    statelink_mask = statelink_info.get('netmask', None)
                    statelink_standby = statelink_info.get('ip_addr_standby', None)
                    
                    self.config_statelink(
                        unit=self.primary,
                        link_name=state_link_name,
                        intf=statelink_intf,
                        ip_addr=statelink_ip,
                        ip_netmask=statelink_mask,
                        ip_addr_standby=statelink_standby
                    )
                else:
                    self.config_statelink(
                        unit=self.primary,
                        link_name=folink_name,
                    )

            # Configure secondary unit
            self.config_folink(
                self.secondary, folink_intf, folink_ip, folink_mask, folink_standby)
            
            if 'state_link' in self.secondary.topo.failover:
                if statelink_intf != folink_intf:
                    cmd = 'interface {}'.format(statelink_intf)
                    cmd += '\nno shutdown'
                    self.secondary.config(cmd, exception_on_bad_config=True)
            self.config_as_secondary(self.secondary)

            # Enable HA
            self.enable_failover(self.secondary, check_status=False)
            self.enable_failover(self.primary, check_status=True)

        # Check HA status
        try:
            self.update_failover_roles()
        except RuntimeError:
            warning_msg = """
            \nWARNING:    Unable to determine primary/secondary or active/standby roles. 
            This will impact further operations defined in this module.
            Double check your configuration and use the config methods provided 
            in this module if something is missed. And then run update_failover_roles again.\n
            """
            self.primary.logger.warning(warning_msg)
            pass

    @staticmethod
    def config_as_primary(unit):
        """Configure a unit to be primary

        :param unit: ASA instance to be configured as primary
        :return: None

        """
        cmd = 'failover lan unit primary'
        unit.config(cmd, exception_on_bad_config=True)
    
    @staticmethod
    def config_as_secondary(unit):
        """Configure a unit to be secondary

        :param unit: ASA instance to be configured as secondary
        :return: None

        """
        cmd = 'failover lan unit secondary'
        unit.config(cmd, exception_on_bad_config=True)

    @staticmethod
    def config_folink(unit, intf, ip_addr, ip_netmask, ip_addr_standby, link_name=None):
        """Configure HA failover link

        :param unit: unit that you want to apply the config
        :param intf: physical interface of failover link
        :param ip_addr: Primary IP address of failover link
        :param ip_netmask: Netmask of failover IP
        :param ip_addr_standby: Standby IP address of failover link
        :param link_name: Failover link name to be assigned
        :return: None

        """
        link_name = link_name or ASAHALinkNames.DEFAULT_FAILOVER_LINK.value
        cmd = 'interface {intf}'.format(intf=intf)
        cmd += '\nno shutdown'
        cmd += '\nfailover lan interface {link_name} {intf}'.format(link_name=link_name, intf=intf)
        if ':' in ip_addr:
            cmd += '\nfailover interface ip folink {ip}/{mask} standby {standby}'.\
                format(ip=ip_addr, mask=ip_netmask, standby=ip_addr_standby)
        else:
            cmd += '\nfailover interface ip folink {ip} {mask} standby {standby}'.\
                format(ip=ip_addr, mask=ip_netmask, standby=ip_addr_standby)

        unit.config(cmd, exception_on_bad_config=True)

    @staticmethod
    def config_statelink(
            unit, link_name, intf=None, ip_addr=None, ip_netmask=None, ip_addr_standby=None):
        """Configure HA stateful link
        
        :param link_name: Stateful link name to be assigned
        :param unit: unit that you want to apply the config
        :param intf: physical interface of stateful link
        :param ip_addr: Primary IP address of stateful link
        :param ip_netmask: Netmask of stateful IP
        :param ip_addr_standby: Standby IP address of stateful link
        :return: None

        """
        cmd = ''
        intf = intf or ''

        if intf:
            cmd += 'interface {intf}'.format(intf=intf)
            cmd += '\nno shutdown'
        cmd += '\nfailover link {link_name} {intf}'.format(link_name=link_name, intf=intf)
        if ip_addr and ip_netmask and ip_addr_standby:
            if ':' in ip_addr:
                cmd += '\nfailover interface ip statelink {ip}/{mask} standby {standby}'.\
                    format(ip=ip_addr, mask=ip_netmask, standby=ip_addr_standby)
            else:
                cmd += '\nfailover interface ip statelink {ip} {mask} standby {standby}'.\
                    format(ip=ip_addr, mask=ip_netmask, standby=ip_addr_standby)

        unit.config(cmd, exception_on_bad_config=True)

    def is_failover_formed(self):
        """Determine whether an HA pair is formed by checking if both active and standby units
           are actively in the pair.

        :return: True|False
        """

        output = self.primary.execute('show failover | include host')
        if re.search('- ' + ASAHAStates.ACTIVE.value, output) \
            and re.search('- ' + ASAHAStates.STANDBY.value, output):
            return True
        return False

    def enable_failover(self, unit, check_status=False, timeout=300):
        """Enable failover

        :param unit: asa instance where failover will be enabled
        :param check_status: option to check failover status. False by default
        :return: None

        """
        unit.config('failover')

        if check_status:
            dialog = Dialog([
                ['Beginning configuration replication', None, None, True, False],
                ['End Configuration Replication to mate', 'sendline()', None, False, False],
            ])
            dialog.process(unit.asa_conn.spawn_id, timeout=timeout)

            start_time = time.time()
            while time.time() - start_time < timeout:
                if self.is_failover_formed():
                    break
                time.sleep(10)
            else:
                raise RuntimeError('HA pair not formed in {} sec'.format(timeout))

    def switch_active(self, unit, timeout=180):
        """Switch active/standby units

        :param unit: unit to be switched to active
        :param timeout: timeout for units switching
        :return: None

        """
        unit.config('failover active')

        start_time = time.time()
        while time.time() - start_time < timeout:
            if self.is_failover_formed():
                break
            time.sleep(5)
        else:
            raise RuntimeError('HA pair not formed after switchover in {} sec'.format(timeout))

        self.update_failover_roles()

    def update_failover_roles(self):
        """Update failver units roles

        Based on output of show failover, this method determines the roles of primary/secondary
        and active/standby units

        :return: None

        """
        show_failover1 = self.asa_instance1.execute('show failover')

        role1 = re.search('This host: (.*) - (.*)', show_failover1)
        role2 = re.search('Other host: (.*) - (.*)', show_failover1)
        primary_found = secondary_found = False
        active_found = standby_found = False

        if role1 and role2:
            # Determine roles on instance 1
            if role1.group(1).strip() == ASAHAStates.PRIMARY.value:
                primary_found = True
                self.primary = self.asa_instance1
            elif role1.group(1).strip() == ASAHAStates.SECONDARY.value:
                secondary_found = True
                self.secondary = self.asa_instance1

            if role1.group(2).strip() == ASAHAStates.ACTIVE.value:
                active_found = True
                self.active = self.asa_instance1
            elif role1.group(2).strip() == ASAHAStates.STANDBY.value:
                standby_found = True
                self.standby = self.asa_instance1

            # Determine roles on instance 2
            if role2.group(1).strip() == ASAHAStates.PRIMARY.value:
                primary_found = True
                self.primary = self.asa_instance2
            elif role2.group(1).strip() == ASAHAStates.SECONDARY.value:
                secondary_found = True
                self.secondary = self.asa_instance2

            if role2.group(2).strip() == ASAHAStates.ACTIVE.value:
                active_found = True
                self.active = self.asa_instance2
            elif role2.group(2).strip() == ASAHAStates.STANDBY.value:
                standby_found = True
                self.standby = self.asa_instance2
        else:
            raise RuntimeError('Unable to determine units roles')

        if not primary_found:
            raise RuntimeError('Unable to find primary unit')
        if not secondary_found:
            raise RuntimeError('Unable to find secondary unit')
        if not active_found:
            raise RuntimeError('Unable to find active unit')
        if not standby_found:
            raise RuntimeError('Unable to find standby unit')

    def get_dataintf_status(self, nameif):
        """Get status of a non HA interface given its logical name

        :param nameif: logical name of the interface
        :return: link status in both active and standby units in a dict

        """
        status = {}
        active, standby = self.active.execute('show failover').split('Other host')
        for line in active.splitlines():
            match = re.search('Interface {}.*: (.*)'.format(nameif), line)
            if match:
                status['active'] = match.group(1)
                break
        for line in standby.splitlines():
            match = re.search('Interface {}.*: (.*)'.format(nameif), line)
            if match:
                status['standby'] = match.group(1)
                break
        return status

class AsaHAActiveActive(AsaHAActiveStandby):
    """Class ASA active/active HA
    """
    def __init__(self, asa_instance1, asa_instance2, init_ha=True):
        """Initializer of AsaHAActiveStandby.
        By default, instance1 will be primary and instance2 will be secondary.

        :param asa_instance1: connection instance1 that created by AsaConfig
        :param asa_instance2: connection instance2 that created by AsaConfig
        :return: None

        """
        self.asa_instance1 = asa_instance1
        self.asa_instance2 = asa_instance2
        self.primary = asa_instance1
        self.secondary = asa_instance2

        if init_ha:
            # Configure primary unit
            self.config_as_primary(self.primary)

            folink_info = self.primary.topo.failover.failover_link
            if 'vlan' in folink_info and folink_info.vlan:
                folink_intf = '{hardware}.{vlan}'.format(hardware=folink_info.interface, vlan=folink_info.vlan)
                cmd = "interface %s\nno shutdown\n" % folink_info.interface
                cmd += "interface %s\nvlan %s" % (folink_intf, folink_info.vlan)
                self.primary.config(cmd, ctx = 'system')
                self.secondary.config(cmd, ctx='system')
            else:
                folink_intf = '{hardware}'.format(hardware=folink_info.interface)

            folink_ip = folink_info.ip_addr
            folink_mask = folink_info.netmask
            folink_standby = folink_info.ip_addr_standby
            folink_name = folink_info.get('link_name', ASAHALinkNames.DEFAULT_FAILOVER_LINK.value)
            self.config_folink(self.primary, folink_intf, folink_ip, folink_mask, folink_standby, folink_name)

            # Stateful link is optional.
            # Check existence in topo first if 'state_link' in self.primary.topo.failover:
            if 'state_link' in self.primary.topo.failover:
                statelink_info = self.primary.topo.failover.state_link
                if 'vlan' in statelink_info and statelink_info.vlan:
                    statelink_intf = '{hardware}.{vlan}'.format(hardware=statelink_info.interface, vlan=statelink_info.vlan)
                    cmd = "interface %s\nno shutdown\n" % statelink_info.interface
                    cmd += "interface %s\nvlan %s" % (statelink_intf, statelink_info.vlan)
                    self.primary.config(cmd, ctx='system')
                    self.secondary.config(cmd, ctx='system')
                else:
                    statelink_intf = '{hardware}'.format(hardware=statelink_info.interface)

                if statelink_intf != folink_intf:
                    state_link_name = statelink_info.get(
                        'link_name', ASAHALinkNames.DEFAULT_STATE_LINK.value)
                    statelink_ip = statelink_info.get('ip_addr', None)
                    statelink_mask = statelink_info.get('netmask', None)
                    statelink_standby = statelink_info.get('ip_addr_standby', None)

                    self.config_statelink(
                        unit=self.primary,
                        link_name=state_link_name,
                        intf=statelink_intf,
                        ip_addr=statelink_ip,
                        ip_netmask=statelink_mask,
                        ip_addr_standby=statelink_standby
                    )
                else:
                    self.config_statelink(
                        unit=self.primary,
                        link_name=folink_name,
                    )

            # Configure secondary unit
            self.config_folink(
                self.secondary, folink_intf, folink_ip, folink_mask, folink_standby)

            if 'state_link' in self.secondary.topo.failover:
                if statelink_intf != folink_intf:
                    cmd = 'interface {}'.format(statelink_intf)
                    cmd += '\nno shutdown'
                    self.secondary.config(cmd, exception_on_bad_config=True)
            self.config_as_secondary(self.secondary)

            # Configure failover groups for active/active
            self.config_failover_groups(self.primary)
            self.config_failover_groups(self.secondary)

            # Enable HA
            self.enable_failover(self.secondary, check_status=False)
            self.enable_failover(self.primary, check_status=True)

            # Join failover groups for each context configured in failover_groups
            for fo_grp in self.primary.topo.failover_groups:
                self.join_failover_group(unit=self.primary, context=fo_grp.ctx, failover_group_idx=fo_grp.idx)

        # Check HA status
        try:
            self.update_failover_roles()
        except RuntimeError:
            warning_msg = """
            \nWARNING:    Unable to determine primary/secondary or active/standby roles. 
            This will impact further operations defined in this module.
            Double check your configuration and use the config methods provided 
            in this module if something is missed. And then run update_failover_roles again.\n
            """
            self.primary.logger.warning(warning_msg)
            pass

    def config_failover_groups(self, unit):
        """Configure 2 failover groups on primary or secondary in active/active configuration

        :param unit: ASA instance to be configured
        :return: None

        """
        cmd = 'failover group 1'
        cmd += '\nprimary'
        cmd += '\npreempt'
        cmd += '\nexit'
        cmd += '\nfailover group 2'
        cmd += '\nsecondary'
        cmd += '\npreempt'
        cmd += '\nexit'
        unit.config(cmd, exception_on_bad_config=True)

    def join_failover_group(self, unit, context, failover_group_idx):
        """Configure to join a failover group for a context in active/active configuration

        :param unit: ASA instance to be configured
        :param context: ASA context
        :param failover_group_idx: 1 or 2 as an example
        :return: None

        """
        unit.config('no failover', ctx='system')
        unit.config('context {}'.format(context), ctx='system')
        unit.config('join-failover-group {}'.format(failover_group_idx))
        unit.config('exit')
        self.enable_failover(unit, check_status=True)

    def is_failover_formed(self):
        """Determine whether an HA pair is formed by checking if both primary and secondary
           are actively in the pair in active/active mode

        :return: True|False
        """

        self.primary.change_to_context(ctx='system')
        output = self.primary.execute('show failover | include Group 1')
        if not re.search(ASAHAStates.ACTIVE.value, output) \
            or not re.search(ASAHAStates.STANDBY.value, output):
            return False
        output = self.primary.execute('show failover | include Group 2')
        if not re.search(ASAHAStates.ACTIVE.value, output) \
            or not re.search(ASAHAStates.STANDBY.value, output):
            return False
        return True

    def validate_failover_groups(self):
        """Determine whether failover groups are formed
           in active/active configuration

        :return: True|False
        """

        flag = False
        unit = self.primary
        unit.change_to_context('system')
        unit.execute('show failover group 1')
        unit.execute('show failover group 2')
        output = unit.execute('show failover group 1 | include State')
        if re.search(ASAHAStates.ACTIVE.value, output) \
            and re.search(ASAHAStates.STANDBY.value, output):
            flag = True
        output = unit.execute('show failover group 2 | include State')
        if re.search(ASAHAStates.ACTIVE.value, output) \
            and re.search(ASAHAStates.STANDBY.value, output):
            flag = True
        return flag

    def update_failover_roles(self):
        """Update failver units roles

        Based on output of show failover, this method determines the roles of primary/secondary
        active/active

        :return: None

        """
        show_failover1 = self.asa_instance1.execute('show failover')

        role1 = re.search('This host: (.*)', show_failover1)
        role2 = re.search('Other host: (.*)', show_failover1)
        primary_found = secondary_found = False

        if role1 and role2:
            # Determine roles on instance 1
            if role1.group(1).strip() == ASAHAStates.PRIMARY.value:
                primary_found = True
                self.primary = self.asa_instance1
            elif role1.group(1).strip() == ASAHAStates.SECONDARY.value:
                secondary_found = True
                self.secondary = self.asa_instance1

            # Determine roles on instance 2
            if role2.group(1).strip() == ASAHAStates.PRIMARY.value:
                primary_found = True
                self.primary = self.asa_instance2
            elif role2.group(1).strip() == ASAHAStates.SECONDARY.value:
                secondary_found = True
                self.secondary = self.asa_instance2

        else:
            raise RuntimeError('Unable to determine units roles')

        if not primary_found:
            raise RuntimeError('Unable to find primary unit')
        if not secondary_found:
            raise RuntimeError('Unable to find secondary unit')





import datetime
import logging
import re
import time

from unicon.eal.dialogs import Dialog

logger = logging.getLogger(__name__)

from unicon.eal.expect import TimeoutError
from ...general.actions.basic import BasicDevice, BasicLine

try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric
from ...general.actions.power_bar import power_cycle_all_ports

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass

MAX_RETRY_COUNT = 3


class Elektra(BasicDevice):
    def __init__(self, hostname='firepower', login_password='Admin123',
                 sudo_password='Admin123', enable_password='', *args, **kwargs):
        """Constructor for Elektra.

        :param hostname: hostname for the device
        :param login_password: device login password with user name 'admin'
        :param sudo_password: device sudo password for 'root'
        :param enable_password: device enable password
        Add *args and **kwargs so that the initiator can be invoked with
        additional (unused) arguments.

        """

        super().__init__()
        publish_kick_metric('device.elektra.init', 1)

        from .patterns import ElektraPatterns
        self.patterns = ElektraPatterns(hostname, login_password, sudo_password, enable_password)

        # create the state machine that contains the proper attributes.
        from .statemachine import ElektraStatemachine
        self.sm = ElektraStatemachine(self.patterns)

        # important: set self.line_class so that a proper line can be created
        # by ssh_console(), etc.
        self.line_class = ElektraLine

    def log_checks(self, ftd_line, list_files=['/var/log/messages'],
                   search_strings=['fatal', 'error'], exclude_strings=[], timeout=300):
        """Wrapper function to Get logs for FTD.

        :param ftd_line: Instance of FTD line used to connect to FTD
        :param list_files: List of file paths for files to search in
        :param search_strings: List of keywords to be searched in the logs
        :param exclude_strings: List of keywords to be excluded in the logs
        :param timeout: The number for seconds to wait for log retrieval
        
        e.g list_files = ['/var/log/boot_1341232','/var/log/boot_*']
            search_strings = ['fatal','error', 'crash']
            exclude_strings = ['ssl_flow_errors', 'firstboot.S09']

        """

        self.sm.go_to('sudo_state', ftd_line.spawn_id)

        grep_command_list = []
        exclude_line = ''

        if exclude_strings:
            exclude_cmd = ['| grep -v {}'.format(string) for string in exclude_strings]
            exclude_line = ''.join(exclude_cmd)

        if list_files and search_strings:
            for file in list_files:
                for string in search_strings:
                    grep_command = "grep -Ii {} {} | " \
                                   "sort -u {}".format(string, file, exclude_line)
                    grep_command_list.append(grep_command)
        try:
            output_log = ftd_line.execute_lines(("\n".join(grep_command_list)),
                                                timeout=timeout)
        except:
            output_log = "Log retrieval command timed out"

        logger.info("""
            ***********************************************************

            Logs for the requested files in FTD are : -

            {}

            ***********************************************************
            """.format(output_log))


class ElektraLine(BasicLine):

    def go_to(self, state, timeout=10):
        """
            Override parent go_to function to enable hop_wise flag.
        """
        super().go_to(state, hop_wise=True, timeout=timeout)

    def change_to_context(self, ctx, timeout=None):
        """Multi-context support.

        :param ctx: context name
        :param timeout: timeout to change context
        :return: None

        """

        # always raise exception if specified context is invalid
        self.execute('changeto context {}'.format(ctx), timeout=timeout,
                     exception_on_bad_command=True)

    def change_to_system(self, timeout=None):
        """Multi-context support, change to system context.

        :param timeout: timeout to change context
        :return None

        """

        self.execute('changeto system', timeout=timeout,
                     exception_on_bad_command=True)

    def find_matched_str(self, expected_str, return_data, matched_whole_line=False):
        """find matched line or substring from return_data.

        :param expected_str: expected string for the output of cmd executed
        :param return_data: command output
        :param matched_whole_line: True if want to match the whole line,
                                    otherwise, match the substring
        :return matched_obj: the matched object

        """

        matched_obj = False
        if matched_whole_line:
            for line in return_data.split('\n'):
                line = line.strip()
                if line == '':
                    continue
                logger.debug('line=' + line)
                if line == expected_str:
                    matched_obj = line
                    break
        else:
            matched_obj = re.search(expected_str, return_data)
        return matched_obj

    def execute_and_verify(self, cmd,
                           expected_str,
                           cmd_set_config='',
                           timeout=60,
                           interval=2,
                           timeout_total=60,
                           retry_count=MAX_RETRY_COUNT,
                           matched_whole_line=False):
        """Execute a command, verify the output, set the config if the output
        is not expected, The command will be run and retried for no more than
        retry_count, and no more than timeout_total in seconds Assert if
        expected_str is not found in the output of cmd.

        :param cmd: command to be executed
        :param expected_str: expected string for the output of cmd executed
        :param cmd_set_config: command to be executed if expected_str is not seen,
                                if cmd_set_config is an empty string, the function is waiting only
        :param timeout: timeout for execute() in seconds
        :param interval: wait time for each retry in seconds
        :param timeout_total: total wait time in seconds
        :param retry_count: Max count for retries
        :param matched_whole_line: True if want to match the whole line
        :return: None

        """

        logger.info("""Command: {}
            Expected String: {}
            Command to set configuration: {}
        """.format(cmd, expected_str, cmd_set_config))
        count = 0
        expected = False
        start_time = datetime.datetime.now()
        elapsed_time = 0
        while count < retry_count and elapsed_time < timeout_total:
            return_data = self.execute(cmd, timeout=timeout)
            print('>>>' + return_data + '<<<')
            logger.info('command output: {}'.format(return_data))
            matched_obj = self.find_matched_str(expected_str, return_data,
                                                matched_whole_line)
            if matched_obj is None:
                if cmd_set_config is not '':
                    # Execute the command to get expected_str
                    self.execute(cmd_set_config, timeout=timeout)
                    self.execute('commit', timeout=timeout)
                count += 1
                now = datetime.datetime.now()
                elapsed_time = (now - start_time).total_seconds()
                logger.info('time elapsed: {} seconds'.format(round(elapsed_time)))
                time.sleep(interval)
            else:
                logger.info('found expected string: {}'.format(expected_str))
                expected = True
                break
        assert expected, "Command: {}, Expected String: {}" \
                          " not found".format(cmd, expected_str)

    def configure_terminal(self):
        """Configure terminal settings.

        :return None

        """

        self.go_to('enable_state')
        self.execute('terminal pager 0')
        self.go_to('config_state')
        self.execute('ssh timeout 60')
        self.execute('terminal width 511')
        self.go_to('enable_state')

    def download_and_set_boot_module(self):
        """Download boot image, and set module to downloaded image.

        :return: None

        """

        asa_boot_image = 'disk0:asaboot.img'
        logger.info('=== Download boot image {} to {}'.format(self.boot_image,
                                                              asa_boot_image))
        command = 'copy /noconfirm {} {}'.format(self.boot_image, asa_boot_image)
        self.execute_and_verify(cmd=command,
                                expected_str='bytes copied in',
                                timeout=120,
                                timeout_total=360,
                                retry_count=self.retry_count)
        logger.info("=== Stop the current recovery process if any")
        output = self.execute('sw-module module sfr recover stop')
        if 'Module sfr cannot have recovery stopped, not in recover state' not in output:
            logger.info('=== Check that current sfr recovery has stopped')
            self.execute_and_verify(cmd='show module sfr details | grep ^Status',
                                    expected_str='    Unresponsive   No Image Present',
                                    timeout=120,
                                    interval=10,
                                    timeout_total=360,
                                    retry_count=50)
        logger.info('=== Set module to {}'.format(asa_boot_image))
        self.execute('sw-module module sfr recover '
                     'configure image {}'.format(asa_boot_image))

        self.spawn_id.sendline('sw-module module sfr recover boot')
        d1 = Dialog([
            ['Recover module sfr', 'sendline()', None, False, False],
        ])
        d1.process(self.spawn_id, timeout=30)

        logger.info('=== Check module sfr status')
        # if sfr status is 'Up', the module recover command has to be reissued
        output = self.execute("show module sfr details | grep ^Status", timeout=120)
        if any(state in output for state in ["Init", "Up"]):
            self.execute_and_verify(cmd='show module sfr details | grep ^Status',
                                    expected_str='    Up',
                                    timeout=120,
                                    interval=10,
                                    timeout_total=360,
                                    retry_count=50)
            self.spawn_id.sendline('sw-module module sfr recover boot')
            d1.process(self.spawn_id, timeout=30)
        # Expect 'Recover', otherwise it is 'Shutting Down'
        self.execute_and_verify(cmd='show module sfr details | grep ^Status',
                                expected_str='    Recover',
                                timeout=120,
                                interval=10,
                                timeout_total=360,
                                retry_count=50)
        self.execute_and_verify(cmd='show module sfr details | grep ^Console session:',
                                expected_str='    Ready',
                                timeout=120,
                                interval=10,
                                timeout_total=360,
                                retry_count=50)

        logger.info('=== Connect sfr console')
        self.spawn_id.sendline()
        # Wait for this error state to be finished:
        # 'ERROR: Failed opening console session with
        # module sfr. Module is in "Recover" state.'
        # 'Please try again later.'
        self.spawn_id.sendline('session sfr console')
        d2 = Dialog([
            ['({}|firepower)#'.format(self.sm.patterns.hostname),
             'sendline({})'.format('session sfr console'), None, True, False],
            ['Escape character sequence is', 'sendline()',
             None, True, False],
            [r'\w+ login:', 'sendline({})'.format(self.sm.patterns.username),
             None, True, False],
            ['Password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['-boot>', 'sendline()', None, False, False],
        ])
        d2.process(self.spawn_id, timeout=600)

    def boot_configure(self):
        """In boot CLI mode, set device network info, DNS server and domain.

        :return: None

        """

        self.spawn_id.sendline('setup')
        self.spawn_id.expect('-boot>')
        set_ntp_server = 'n' if self.ntp_server == None else 'y'
        d2 = Dialog([
            ['Enter a hostname ', 'sendline({})'.format(self.hostname),
             None, True, False],
            ['Do you want to configure IPv4 address on management interface',
             'sendline(y)', None, True, False],
            ['Do you want to enable DHCP for IPv4 address assignment on '
             'management interface', 'sendline(n)', None, True, False],
            ['Enter an IPv4 address', 'sendline({})'.format(self.uut_ip),
             None, True, False],
            ['Enter the netmask', 'sendline({})'.format(self.uut_netmask),
             None, True, False],
            ['Enter the gateway', 'sendline({})'.format(self.uut_gateway),
             None, True, False],
            ['Do you want to configure static IPv6 address on management '
             'interface', 'sendline(n)', None, True, False],
            ['Enter the primary DNS server IP address',
             'sendline({})'.format(self.dns_server), None, True, False],
            ['Do you want to configure Secondary DNS Server', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Local Domain Name', 'sendline(n)',
             None, True, False],
            ['Do you want to configure Search domains', 'sendline(y)', None,
             True, False],
            ['Enter the comma separated list for search domains',
             'sendline({})'.format(self.search_domains), None, True, False],
            ['Do you want to enable the NTP service', 'sendline({})'.format(set_ntp_server), None,
             True, False],
            ['Enter the NTP servers separated by commas', 'sendline({})'.format(self.ntp_server), None,
             True, False],
            ['Do you want to enable the NTP symmetric key authentication?', 'sendline(n)', None, True, False],
            ['Apply the changes', 'sendline(y)', None, True, False],
            ['Press ENTER to continue...', 'sendline()', None, False, False],
        ])
        d2.process(self.spawn_id, timeout=120)

    def ftd_install(self):
        """Perform ping test and verify the network connectivity to http server
        Install SFR pkg image Enter device network info, hostname.

        :return: None

        """

        self.spawn_id.sendline('')
        self.spawn_id.expect('-boot>', timeout=20)
      
        for i in range(20):
            self.spawn_id.sendline('ping -c 1 {}'.format(self.http_server))
            try:
                self.spawn_id.expect('64 bytes from', timeout=5)
            except TimeoutError:
                continue
            else:
                break
        else:
            raise RuntimeError(">>>>>> Ping not working")

        d1 = Dialog([
            ['-boot>', 'sendline(system install {})'.format(self.pkg_image),
                None, True, False],
            ['Do you want to continue?', 'sendline(y)', None, True, False],
            ['Upgrade aborted', 'sendline()', None, False, False],
        ])
        count = 0
        while count < self.retry_count:
            try:
                d1.process(self.spawn_id, timeout=300)
                count += 1
                logger.info('captured "Upgrade aborted": {} times'.format(count))
                time.sleep(5)
            except:
                logger.info('did not capture Upgrade aborted within 5 minutes')
                break
        assert count < self.retry_count, 'ftd installation failed' \
            ', please check ftd package url: "{}"'.format(self.pkg_image)

        d2 = Dialog([
            ['Do you want to continue with upgrade?', 'sendline(y)', None,
             True, True],
            ["Press 'Enter' to reboot the system.", 'sendline()', None, True,
             True],
            ['The system is going down for reboot NOW', 'sendline()',
              None, False, True],
            #['{}# '.format(self.hostname), 'sendline()', None, False, True],
        ])
        d2.process(self.spawn_id, timeout=2400)

    def ftd_configure(self):
        """Accept EULA, enter new password and configure network for ftd.

        :return: None

        """

        time.sleep(10)
        for i in range(20):
            self.spawn_id.sendline('')
            try:
                self.go_to('any')
            except:
                continue
            else:
                break
        else:
            raise RuntimeError(">>>>>> Switching back to enable mode failed")

        self.execute_and_verify(cmd='show module sfr details | grep ^Status',
                                expected_str='Up',
                                timeout=360,
                                interval=10,
                                timeout_total=7200,
                                retry_count=360)

        self.spawn_id.sendline('session sfr console\n')
        d1 = Dialog([
            ['firepower login: ', 'sendline({})'.format(self.sm.patterns.username),
             None, True, False],
            ['Password: ', 'sendline({})'.format(self.sm.patterns.default_password),
             None, True, False],
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True,
             False],
            ['--More--', 'send(q)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ",
             'sendline(YES)', None, False, False],
        ])
        d1.process(self.spawn_id, timeout=600)
        d2 = Dialog([
            ['Enter new password:',
             'sendline({})'.format(self.sm.patterns.login_password),
             None, True, True],
            ['Confirm new password:',
             'sendline({})'.format(self.sm.patterns.login_password),
             None, True, False],
            ['Do you want to configure IPv4', 'sendline(y)', None, True, False]
            ])
        if self.uut_ip6 is None:
            d2.append(['Do you want to configure IPv6', 'sendline(n)', None, True, False])
        else:
            d2.append(['Do you want to configure IPv6', 'sendline(y)', None, True, False])
        d2.append(['Configure IPv4 via DHCP or manually', 'sendline(manual)', None,
                    True, False])
        d2.append(['Enter an IPv4 address for the management interface',
                    'sendline({})'.format(self.uut_ip), None, True, False])
        d2.append(['Enter an IPv4 netmask for the management interface',
                    'sendline({})'.format(self.uut_netmask), None, True, False])
        d2.append(['Enter the IPv4 default gateway for the management interface',
                    'sendline({})'.format(self.uut_gateway), None, True, False])
        if self.uut_ip6 is not None:
            d2.append(['Configure IPv6 via DHCP, router, or manually',
                       'sendline(manual)', None, True, False])
            d2.append(['Enter the IPv6 address for the management interface',
                       'sendline({})'.format(self.uut_ip6), None, True, False])
            d2.append(['Enter the IPv6 address prefix for the management interface',
                       'sendline({})'.format(self.uut_prefix), None, True, False])
            d2.append(['Enter the IPv6 gateway for the management interface',
                       'sendline({})'.format(self.uut_gateway6), None, True, False])
        d2.append(['Enter a fully qualified hostname for this system ',
                    'sendline({})'.format(self.hostname), None, True, False])
        d2.append(['Enter a comma-separated list of DNS servers or',
                    'sendline({})'.format(self.dns_server), None, True, False])
        d2.append(['Enter a comma-separated list of search domains or',
                    'sendline({})'.format(self.search_domains), None, True, False])
        d2.append(['> ', 'sendline()', None, False, False])
        d2.process(self.spawn_id, timeout=900)

        logger.info('fully installed.')

    def validate_version(self):
        """Checks if the installed version matches a the version in package
        url.

        :return: None

        """

        self.go_to('any')
        self.go_to('fireos_state')

        # check version
        response = self.execute('show version', 30)
        #print('response=' +  response + '<')
        build = re.findall('Build\s(\d+)', response)[0]
        version = re.findall(r'(Version\s){1}([0-9.]+\d)', str(response))[0][1]
        ftd_package_file = self.pkg_image.split('/')[-1]
        if build in ftd_package_file and version in ftd_package_file:
            logger.info('>>>>>> show version result:\n{}\nmatches '
                        'ftd package image: {}'.format(response,
                                                       ftd_package_file))
            logger.info('Installed ftd version validated')
        else:
            logger.error('Exception: not the same version and build')
            raise RuntimeError('>>>>>> show version result:\n{}\ndoes not match '
                               'ftd package image: {}'.format(response,
                                                              ftd_package_file))

    def baseline_elektra(self, http_server, boot_image, pkg_image,
                         uut_ip, uut_netmask, uut_gateway, dns_server,
                         hostname='firepower', search_domains='cisco.com',
                         uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                         retry_count=MAX_RETRY_COUNT, ntp_server=None, **kwargs):
        """Baseline device of Elektra, install boot_image, pkg_image and verify
        image installed.

        :param http_server: http server IP
        :param boot_image: in enable mode, download and install boot_image.
                            for example, boot_image='http://10.89.23.80/netboot/ims/Release/6.2.0-362/
                            installers/asasfr-5500x-boot-6.2.0-2.img'
        :param pkg_image: in boot mode, download and install pkg_image.
                            for example, pkg_image='http://10.89.23.80/netboot/ims/Release/6.2.0-362/
                            installers/asasfr-sys-6.2.0-362.pkg'
        :param uut_ip: UUT IP address
        :param uut_netmask: UUT netmask
        :param uut_gateway: UUT gateway
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param dns_server: DNS server
        :param hostname: hostname of the device to be set
        :param search_domains: search domains, defaulted to 'cisco.com'
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param ntp_server: NTP server delimited by comma, defaulted to None,
                            otherwise the value of ntp_server is - e.g. "ntp.esl.cisco.com"
        :return: None

        """

        publish_kick_metric('device.elektra.baseline', 1)

        self.http_server = http_server
        self.boot_image = boot_image
        self.pkg_image = pkg_image
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.retry_count = retry_count
        self.ntp_server = ntp_server

        if self.sm.current_state is not 'boot_state':
            logger.info('=== Configure terminal')
            self.configure_terminal()

            logger.info('=== Download and set module to boot_image')
            self.download_and_set_boot_module()

        logger.info('=== Boot configure')
        self.boot_configure()

        logger.info('=== SFR install: {}'.format(self.pkg_image))
        self.ftd_install()

        logger.info('=== FTD configure')
        self.ftd_configure()

        logger.info('=== Validate version')
        self.validate_version()

    def configure_ssh_v2(self, gateway, netmask, user, passwd, activation_key=''):
        """Function to configure ssh version 2
        
        :param gateway:
        :param netmask: the ip netmask to apply to the ip
        :param user: username of the user
        :param passwd: the password for user
        :param activation_key: string containing a Encryption-3DES-AES activation key
               e.g.: '7717d15c 7451cf52 b81171b4 979c3404 c51501b3'
        :return: None
        
        """

        self.go_to('config_state')
        self.spawn_id.sendline('crypto key generate rsa modulus 1024')
        d = Dialog([
            ['Keypair generation process begin. Please wait', None, None, True, False],
            ['Do you really want to replace them?', 'sendline(no)', None, True, False],
            [self.sm.patterns.prompt.config_prompt, None, None, False, False],
        ])
        d.process(self.spawn_id, 30)

        cmd_lines1 = 'ssh {} {} diagnostic\n' \
                     'ssh timeout 60\n' \
                     'route diagnostic 0 0 {} 1\n' \
                     'username {} password {}\n' \
                     'aaa authentication ssh console LOCAL'.format(gateway, netmask, gateway, user, passwd)

        logger.info("Configure ssh")
        self.config(cmd_lines1)

        if activation_key:
            logger.info("Apply ssh activation key")
            self.config('activation-key {}'.format(activation_key), 300, True)
        try:
            self.config('ssh version 2', None, True)
        except Exception as e:
            logger.info('SSH version2 could not be activated')
            raise e

    def baseline_by_branch_and_version(self, site, branch, version, uut_ip, uut_netmask,
                                       uut_gateway, dns_server, serverIp='', tftpPrefix='', scpPrefix='', docs='', **kwargs):
        """Baseline elektra using branch and version
        
        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: Device IP Address to access TFTP Server
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param \**kwargs:
            :Keyword Arguments, any of below optional parameters:
        :param hostname: hostname of the device to be set; defaulted to firepower
        :param search_domains: search domains, defaulted to 'cisco.com'
        :param retry_count: download retry count, defaulted to MAX_RETRY_COUNT
        :param ntp_server: NTP server delimited by comma, defaulted to None,
                            otherwise the value of ntp_server is - e.g. "ntp.esl.cisco.com"
        :return: None
        """

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            scp_prefix = scpPrefix
            files = docs
        else:
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 'elektra', branch, version)

        try:
            boot_file = [file for file in files if file.endswith('.img')][0]
        except Exception as e:
            raise Exception('Got {} while getting boot image file.'.format(e))
        files.remove(boot_file)
        pkg_file = files[0]

        boot_image = "http://{}/{}/{}".format(server_ip, tftp_prefix[len('asa/'):], boot_file)
        pkg_image = "http://{}/{}/{}".format(server_ip, tftp_prefix[len('asa/'):], pkg_file)

        kwargs['http_server'] = server_ip
        kwargs['boot_image'] = boot_image
        kwargs['pkg_image'] = pkg_image
        kwargs['uut_ip'] = uut_ip
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['dns_server'] = dns_server
        self.baseline_elektra(**kwargs)

    def power_cycle(self, power_bar_server, power_bar_port, power_bar_user='admn', power_bar_pwd='admn'):
        """reboots a device from a Power Data Unit equipment.

        :param power_bar_server: comma-separated string of IP addresses of the PDU's
        :param power_bar_port: comma-separated string of power port on the PDU's
        :param power_bar_user: comma-separated string of usernames for power bar servers
        :param power_bar_pwd: comma-separated string of passwords for power bar servers
        :return: status of power cycle result, True or False

        """

        if power_bar_server == "" or power_bar_port == "":
            return None

        return power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd)



import munch


class M4Patterns:
    """FMC M4 class for device prompt patterns."""
    def __init__(self, hostname, cimc_hostname, login_username, login_password, sudo_password):
        """Constructor for M4Patterns."""

        self.fqdn = hostname
        self.hostname = hostname.split('.')[0]
        self.cimc_hostname = cimc_hostname
        self.login_username = login_username
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.username = 'admin'
        self.default_password = 'Admin123'
        self.prompt = munch.Munch()

        self.prompt.password_prompt = r'.*[Pp]assword: '
        self.prompt.prelogin_prompt = r'\n?\r?({}([\w.]+)?|firepower) login: '.format(self.hostname)
        self.prompt.mio_prompt = r'{}([/\w\-\*\s]+)?# '.format(self.cimc_hostname)
        self.prompt.admin_prompt = r'\r\nadmin@({}|firepower).*$'.format(self.hostname)
        self.prompt.sudo_prompt = r'\r\nroot@({}|firepower).*#'.format(self.hostname)
        self.prompt.fireos_prompt = r'\r\n(\x1bE\x1b\[J)?[\x07]?> '

        self.prompt.lilo_boot_prompt = 'boot:'
        self.prompt.lilo_boot_menu_prompt = 'qqqqqqqqqqqqqqqqqqqqqqqqqqqqqq'
        self.prompt.lilo_os_prompt = "root@\(none\):.*#"



import munch


class EpPatterns:
    """ Ep class for device prompt patterns
    """

    def __init__(self, hostname, login_username, login_password, sudo_password):
        """ Constructor for EpPatterns
        """
        self.hostname = hostname
        self.login_username = login_username
        self.login_password = login_password
        self.sudo_password = sudo_password
        self.prompt = munch.Munch()

        self.prompt.password_prompt = r'.*[Pp]assword.*: '
        self.prompt.prelogin_prompt = r'({} )login: '.format(self.hostname)
        self.prompt.admin_prompt = r'\r\n.*[a-zA-Z0-9.*]+@[a-zA-Z0-9.*].*[$]'
        self.prompt.sudo_prompt = r'\r\n.*[a-zA-Z0-9.*]+@[a-zA-Z0-9.*].*[#]'



import logging
import re

import time
from munch import munchify as bunchify

logger = logging.getLogger(__name__)


def create_dict(data):
    d = bunchify({})
    for line in data.split('\r\n'):
        if ':' in line:
            line = line.split(':')
            if not (line[0].strip() is '' or line[1].strip() is ''):
                if len(line) > 2:
                    d[line[0].strip()] = ':'.join(line[1:]).strip()
                else:
                    d[line[0].strip()] = line[-1].strip()
    return d


def scope(handle, path, conn='console'):
    if path == 'top':
        handle.execute('top')
        handle.execute('')
    else:
        for sc in path.split('/'):
            if conn == 'console':
                handle.execute('scope %s' % sc)
                handle.execute('')


def poll(handle, command, expect, scope='', max_retry=1, interval=1, flags=re.DOTALL):
    if scope != '':
        scope(handle, scope)

    cnt = 0
    while cnt < max_retry:
        cnt += 1
        logger.info("Looking for '%s' in '%s' output" % (expect, command))
        output = handle.execute(command)
        output = re.sub('(\r|\n| )+', ' ', output)
        m = re.search(re.compile(expect, flags), output)
        if not m:
            if cnt == max_retry:
                duration = cnt * interval
                logger.info("'%s' is not found in the output within %s seconds" \
                            % (expect, duration))
                if scope != '':
                    scope(handle, 'top')
                return False
            else:
                time.sleep(interval)
        else:
            duration = cnt * interval
            logger.info("'%s' is found within %s seconds" % (m.group(0), duration))
            if scope != '':
                scope(handle, 'top')
            return True



from unicon.eal.dialogs import Statement
import logging
import time
logger = logging.getLogger(__name__)


def login_handler(spawn, patterns):
    spawn.sendline(patterns.login_username)
    spawn.expect('Password: ')
    spawn.sendline(patterns.login_password)
    r = spawn.expect(['Cisco Firepower', 'Login incorrect'], timeout=10)
    if 'Login incorrect' in r.match_output:
        login_with_factory_default_password(spawn, patterns)

    # if no sleep after login, device will enter a weird state when "connect ftd"
    # is sent, it echoes back "> connect ftd", instead of "> ". What's worse is
    # from here, you can't type anything in. you have to ctrl+u to clear the
    # existing string ("connect ftd") before anything. a sleep will avoid it.
    time.sleep(3)
    spawn.sendline()


def login_with_factory_default_password(spawn, patterns):
    # see CSCvq65263 for details: if device went through an incomplete baseline, it may
    # contain the factory default password "Admin123", which is worth a try for baseline
    # to continue. It's the user's responsibility to not do anything else other than
    # a new baseline.
    logger.debug("Trying factory default password (Admin123)")
    spawn.sendline(patterns.login_username)
    spawn.expect('Password: ')
    spawn.sendline('Admin123')
    spawn.expect('Cisco Firepower')
    logger.debug("factory default password works!")
    logger.debug("Device uses unexpected password of factory default.")
    logger.warning("You need to baseline the device right away.")
    spawn.sendline()


class KpStatements:
    def __init__(self, patterns):
        self.login_password = Statement(pattern=patterns.prompt.prelogin_prompt,
                                        action=login_handler,
                                        args={'patterns': patterns},
                                        loop_continue=False,
                                        continue_timer=True)

        self.login_incorrect = Statement(pattern='Login incorrect',
                                         action=login_with_factory_default_password,
                                         args={'patterns': patterns},
                                         loop_continue=False,
                                         continue_timer=False)



"""
This script is to set the hostname as given in the testbed file to given WM/KP device
Connect to device console (ssh or telnet) as per the connection protocol defined in testbed.
check if FTD(app) is present, if present update the hostname from ftt app, if not change hostname from FXOS itself

"""


import pexpect
import sys
import time
import re
import os
from kick.device2.general.actions.access import clear_line
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
console_handler = logging.StreamHandler()
LEVEL = logging.DEBUG
console_handler = logging.StreamHandler()
kick_logger = logging.getLogger('kick')
kick_logger.setLevel(LEVEL)
kick_logger.addHandler(console_handler)


class ChangeHostname():

    def __init__(self):

        """Constructor of device for hostname change

        :return: None

        """
        pass


    def connect_to_ssh_console_line(self,ts_ip, port, username='vppuser', password='vppuser'):
        """
        get pexpect line.

        ts_ip: terminal server ip (or hostname)
        port: port on terminal server
        """
        clear_line(host=ts_ip, port=int(port) % 100, user=username, pwd=password, access='ssh')
        line = pexpect.spawn("ssh -l {} -p {} {}".format(username, port, ts_ip))
        log_prefix = os.path.splitext(os.path.basename(__file__))[0]
        line.logfile = open('{}.pexpect'.format(log_prefix), 'wb')
        try:
            line.expect("Are you sure you want to continue", timeout=3)
            line.sendline('yes')
        except:
            pass
        line.expect("Password: ")
        line.sendline(password)
        line.expect("Password OK")
        time.sleep(3)
        logger.debug("*******connected to ssh console********")
        return line

    def connect_to_telnet_console_line(self,ts_ip, port, username='vppuser', password='vppuser'):
        """
        get pexpect line.

        ts_ip: terminal server ip (or hostname)
        port: port on terminal server
        """
        clear_line(host=ts_ip, port=int(port) % 100, user=username, pwd=password, access='telnet')
        line = pexpect.spawn("telnet {} {}".format(ts_ip, port))
        log_prefix = os.path.splitext(os.path.basename(__file__))[0]
        line.logfile = open('{}.pexpect'.format(log_prefix), 'wb')

        try:
            line.expect("Connected to.*Escape character is '\^\]'\.", timeout=3)
            line.sendline('')
            logger.debug("*******connected to telnet console without credential********")
        except:
            line.expect("Connected to.*Escape character is '\^\]'\..*Username: ")
            line.sendline(username)
            line.expect("Password: ", timeout=3)
            line.sendline(password)
            line.sendline('')
            logger.debug("*******connected to telnet console with credential********")

        time.sleep(3)
        return line

    def disconnect_from_ssh_console_line(self,line):
        """
        Send disconnect sequence to line.
        """
        # send ctrl+u, \n + ~.
        line.sendcontrol("u")
        line.sendline()
        line.send('~.')
        try:
            line.expect('Connection to .* closed.')
            logger.debug('ssh line disconnected successfully')
        except OSError as e:
            logger.debug('Connection closed message did not appear: '
                            'Encountered exception: {}.'.format(
                traceback.format_tb(e.__traceback__)))
        line.logfile.close()
        logger.info("*******disconnected from ssh console********")

    def drop_to_prelogin(self, line, max_depth=10):
        """
        Drop line (from any other states) to prelogin, by repeatedly doing
        "exit".
        One added benefit of this step is, it forces the transition of
        sudo -> expert -> fireos state. So the future login will be more
        predictable.
        """
        for i in range(max_depth):
            # send ctrl + u, then exit
            # import pdb; pdb.set_trace()
            line.sendcontrol("u")
            time.sleep(1)
            line.sendline("exit")
            time.sleep(1)
            try:
                line.expect(["login: ", "Password: "], timeout=5) # password auth takes long. 5 seconds are needed.
                # logger.debug("DEBUG: after try{}".format(line.before.decode('utf-8')))
                time.sleep(1)
            except:
                pass
            else:
                logger.info("********in prelogin prompt********")
                time.sleep(3)
                break
        else:
            raise RuntimeError("Cannot get to prelogin after 10 exits.")

    def change_hostname_through_ftd(self,line, hostname, username="admin",
                                    password="Ci5c05n0rt!"):
        """
        Connect to ftd, go to sudo state, and change hostname there.
        """
        line.sendline("connect ftd")
        line.expect("> ")
        line.sendline("expert")
        line.expect("admin@.*:/.*\$ ")
        line.sendline("sudo su")
        line.expect("Password: ")
        line.sendline(password)
        line.expect("root@.*:/.*# ")
        line.sendline("hostname {}".format(hostname))
        line.expect("root@.*:/.*# ")
        line.sendline("exit")
        line.expect("admin@.*:/.*\$ ")
        line.sendline("exit")
        line.expect(">")
        line.sendline("exit")

    def change_hostname_through_scope_system(self,line, hostname):
        """
        Change hostname from scope system. This only works if no ftd app is present.
        """
        line.sendline("scope system\nset name {}\ncommit-buffer\nexit".format(hostname))

    def from_prelogin_to_logged_in(self,line, username, password):
        """
        Move from prelogin state to logged_in.
        """
        # import pdb; pdb.set_trace()
        line.sendcontrol("u")
        time.sleep(1)
        line.sendline()
        line.expect("login: ")
        line.sendline(username)
        line.expect("Password: ")
        line.sendline(password)
        line.expect("Last login: ")
        line.expect("[\r\n]+\S+# ")

    def from_logged_in_to_rommon(self,line):

        """
                Move from prelogin state to logged_in as this is the only way to recover device from "failed" state
        """
        line.sendline()
        line.sendline("connect local-mgmt")
        line.sendline("format everything")
        line.expect("Do you still want to format")
        line.sendline("yes")
        line.expect("Boot in 10 seconds.", timeout=180)
        line.send(chr(27))
        line.expect("[\r\n]+rommon.*> ")

    def change_hostname_from_prelogin(self,line, hostname, username="admin",
                                      password="Ci5c05n0rt!"):

        # go to chassis prompt
        self.from_prelogin_to_logged_in(line, username, password)

        # check if ftd is present
        line.sendline("top\nscope ssa\nshow app\ntop")
        try:
            line.expect("non-existing", timeout=5)
        except:
            pass
        with_ftd = self.check_if_ftd_is_present(line.before.decode('utf-8'))

        if with_ftd:
            line.sendline("connect ftd")
            # in case ftd is in sudo/expert state, force it to go through multiple
            # "exit" sequences, so that device ends up in fireos state.
            self.drop_to_prelogin(line)
            self.from_prelogin_to_logged_in(line, username, password)
            self.change_hostname_through_ftd(line, hostname, username, password)
        else:
            self.change_hostname_through_scope_system(line, hostname)

        line.expect("{}# ".format(hostname))

        logger.info("********hostname changed********")

    def check_if_ftd_is_present(self,show_app_output):
        """
        Check if there is an ftd app present.
        """
        r = re.search("ftd\s+[\d\.]+[^\r\n]+Application", show_app_output)
        if r:
            logger.info("*********ftd app found**********")
            return True
        else:
            logger.info("*********ftd app NOT found**********")
            return False





"""Library to load new images on a Sensor Series3.

We require a file server to be co-located with the device. This file
server has both the img files mounted.  Image downloading is via http.

The file server is a PXE server in each site.

"""
import logging
import os
import subprocess
import re
import urllib
import time

import datetime

logger = logging.getLogger(__name__)

from unicon.eal.expect import Spawn, TimeoutError
from unicon.eal.dialogs import Dialog
from kick.device2.series3.actions.webserver import Webserver
from ...general.actions.basic import BasicDevice, BasicLine
from kick.device2.general.actions.power_bar import power_cycle_all_ports
try:
    from kick.graphite.graphite import publish_kick_metric
except ImportError:
    from kick.metrics.metrics import publish_kick_metric

KICK_EXTERNAL = False

try:
    from kick.file_servers.file_servers import *
except ImportError:
    KICK_EXTERNAL = True
    pass


DUT_DIR_TMP = '/tmp'
DUT_DIR_BOOT = '/boot'
DUT_DIR_ETC = '/etc'
DUT_LILO_TMPL_FILE = 'lilo.conf'
DUT_LILO_FILE = 'lilo.conf.atf'
DUT_LILO_TMPL_PATH = '{}/{}'.format(DUT_DIR_ETC, DUT_LILO_TMPL_FILE)
DUT_LILO_TMP_PATH = '{}/{}'.format(DUT_DIR_TMP, DUT_LILO_FILE)
DUT_LILO_PATH = '{}/{}'.format(DUT_DIR_ETC, DUT_LILO_FILE)
MAX_TIME_INSTALL = 7200
DEFAULT_TIMEOUT = 10
SSH_DEFAULT_TIMEOUT = 90


class Series3(BasicDevice):
    def __init__(self, hostname='firepower', login_username='admin',
                 login_password='Admin123',
                 sudo_password='Admin123'):
        """Constructor for FTD.

        :param hostname: hostname for the device
        :param login_username: user credentials
        :param login_password: device login password with user name 'admin'
        :param sudo_password: device sudo password for 'root'

        """

        super().__init__()
        publish_kick_metric('device.series3.init', 1)
        from .patterns import Series3Patterns
        self.patterns = Series3Patterns(hostname, login_username,
                                        login_password, sudo_password)

        # create the state machine that contains the proper attributes.
        from .statemachine import Series3Statemachine
        self.sm = Series3Statemachine(self.patterns)

        self.line_class = Series3Line

    def ssh_vty(self, ip, port, username='admin', password='Admin123', timeout=None,
                line_type='ssh', rsa_key=None):
        """
        Set up an ssh connection to device's interface.

        This goes into device's ip address, not console.

         :param ip: ip address of terminal server
         :param port: port of device on terminal server
         :param username: usually "admin"
         :param password: usually "Admin123"
         :param line_type: 'ssh'
         :param timeout: in seconds
         :param rsa_key: identity file (full path)
         :return: a line object (where users can call execute(), for example)

         """

        publish_kick_metric('device.series3.ssh', 1)
        if not timeout:
            timeout = self.default_timeout

        from .constants import Series3Constants
        Series3Constants.uut_ssh_ip = ip
        Series3Constants.uut_ssh_port = port
        
        if rsa_key:
            resp = subprocess.getoutput('chmod 400 {}'.format(rsa_key))
            if 'No such file or directory' in resp:
                raise RuntimeError(
                    'The identity file {} you provided does not exist'.format(
                        rsa_key))
            spawn = Spawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                          '-i {} -l {} -p {} {} \n'.format(rsa_key, username, port, ip))
        else:
            spawn = Spawn('ssh -o UserKnownHostsFile=/dev/null -o StrictHostKeyChecking=no '
                          '-l {} -p {} {} \n'.format(username, port, ip))
        d1 = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(password),
             None, False, False],
        ])
        try:
            d1.process(spawn, timeout=timeout)
        # spawn.sendline()
        except TimeoutError:
            logger.info("\n SSH Connection not available...")
            pass

        ssh_line = Series3Line(spawn, self.sm, line_type, timeout=timeout)
        return ssh_line

    def log_checks(self, ftd_line, list_files=['/var/log/messages'],
                   search_strings=['fatal', 'error'], exclude_strings=[], timeout=300):
        """Wrapper function to Get logs for FTD.

        :param ftd_line: Instance of FTD line used to connect to FTD
        :param list_files: List of file paths for files to search in
        :param search_strings: List of keywords to be searched in the logs
        :param exclude_strings: List of keywords to be excluded in the logs
        :param timeout: The number for seconds to wait for log retrieval
                        e.g list_files = ['/var/log/boot_1341232','/var/log/boot_*']
                        search_strings = ['fatal','error', 'crash']
                        exclude_strings = ['ssl_flow_errors', 'firstboot.S09']

        """

        self.sm.go_to('sudo_state', ftd_line.spawn_id)

        grep_command_list = []
        exclude_line = ''

        if exclude_strings:
            exclude_cmd = ['| grep -v {}'.format(string) for string in exclude_strings]
            exclude_line = ''.join(exclude_cmd)

        if list_files and search_strings:
            for file in list_files:
                for string in search_strings:
                    grep_command = "grep -Ii {} {} | " \
                                   "sort -u {}".format(string, file, exclude_line)
                    grep_command_list.append(grep_command)
        try:
            output_log = ftd_line.execute_lines(("\n".join(grep_command_list)),
                                                timeout=timeout)
        except:
            output_log = "Log retrieval command timed out"

        logger.info("""
            ***********************************************************

            Logs for the requested files in FTD are : -

            {}

            ***********************************************************
            """.format(output_log))


class Series3Line(BasicLine):
    def __init__(self, spawn, sm, type, timeout=None):
        """Constructor for series3 line A

        :param spawn: spawn of the line
        :param sm: object of statemachine
        :param type: line type, e.g. 'ssh', 'telnet'
        :return: None

        """
        self.spawn_id = spawn
        self.sm = sm
        self.type = type
        self.set_default_timeout(DEFAULT_TIMEOUT)
        try:
            sm.go_to('any', spawn, timeout=timeout)
        except:
            # Catch the case ftd is back from installation
            # Does not have a stable state
            pass

    def expert_execute(self, cmd, timeout=None, exception_on_bad_command=False):
        """
        :param cmd: cpmmand to be executed in expert state given as a string
        :param timeout: in seconds
        :param exception_on_bad_command: True/False - whether to raise an exception
            on a bad command
        :return: output as string
        """
        self.go_to('expert_state')
        return self.execute(cmd, timeout, exception_on_bad_command)

    ### the following functions are for device baselining. This only works on
    ### a ssh connection.
    def generate_config_file(self):
        """Compose the config file. Copy the config file to
        self.http_dir_for_build on Webserver. Create symbolic link to the iso
        image on Webserver.

        :return: None

        """

        server = Webserver(hostname=self.scp_hostname,
                           login_username=self.scp_username,
                           login_password=self.scp_password)
        dev = server.ssh_vty(self.scp_server)
        output = dev.execute('mkdir -p -m 775 {}'.format(self.http_dir_for_build))
        result = re.search('File exists', output)
        if result is not None:
            logger.info('File exists and it will be deleted ...')
            dev.execute('rm -rf {}'.format(self.http_dir_for_build))
            dev.execute('mkdir -p -m 775 {}'.format(self.http_dir_for_build))
        dev.execute('ln -s {}/{} {}'.format(self.http_dir_root,
                                            self.iso_image_path, self.http_dir_for_build))
        config_file_content = """TRANS=httpsrv
SRV={}
IPCONF=dhcp
REPLACE_TFTP_BOOT_FILE=Yes
TFTPSERVER={}
FN={}
TARGET_ARCH={}
""".format(self.http_server,
           self.http_server,
           self.iso_image_path,
           self.arch)
        dev.execute('echo "{}" > {}'.format(config_file_content, self.config_path))
        output = dev.execute('ls {}'.format(self.config_path))
        output = dev.execute('cat {}'.format(self.config_path))
        output = dev.execute('wget -O - http://{}:{}/{}'.format(self.http_server, self.http_port, self.config_file_url))
        assert 'TRANS=httpsrv' in output, \
            '\nFailed to access the boot configuration file.\nVerify http_dir_tmp is correct.\nhttp://{}:{}/{}'. \
                format(self.http_server, self.http_port, self.config_file_url)
        dev.disconnect()

    def scp_file_to_local(self, local_dir, sftp_ip, sftp_port,
                          username, password, remote_path):
        """Scp a remote file to DUT.

        :param local_dir: local directory as the destination directory on DUT
        :param sftp_ip: remote sftp server ip address
        :param sftp_port: remote sftp server port
        :param username: username to login to sftp server
        :param password: password to login to sftp server
        :param remote_path: file path with filename on remote sftp server
        :return: None

        """

        filename = remote_path.split('/')[-1]
        self.execute('rm -f {}/{}'.format(local_dir, filename))
        self.spawn_id.sendline('scp -P {} {}@{}:{} {}'.format(sftp_port,
                                                              username, sftp_ip, remote_path, local_dir))
        current_prompt = self.sm.get_state(self.sm.current_state).pattern
        d1 = Dialog([
            ['continue connecting (yes/no)?', 'sendline(yes)', None, True, False],
            ['(p|P)assword:', 'sendline({})'.format(password), None, True, False],
            [current_prompt, 'sendline()', None, False, False],
        ])
        d1.process(self.spawn_id, timeout=300)
        self.execute('pwd', timeout=180)
        time.sleep(1)
        self.read_buffer()
        output = self.execute('ls -ltr {}/{}'.format(local_dir, filename),
                              timeout=20)
        full_filename = output.split('/')[-1]
        logger.info('filename={}, full_filename={}'.format(filename,
                                                           full_filename))
        assert ('No such file or directory' not in output) and \
               (filename[:-1] in full_filename), \
            'Failed to copy {}'.format(filename)
        return full_filename

    def copy_bz_usb_files_setup_lilo_to_install(self):
        """Scp bz and usb image files from remote sftp server to DUT. Configure
        lilo file on DUT. Execute run_lilo_and_depmod. Mount usb. Reboot to
        install.

        :return: None

        """

        # Scp bz and usb files
        self.go_to('sudo_state')

        full_bz_filename = self.scp_file_to_local(local_dir=DUT_DIR_TMP,
                                                  sftp_ip=self.scp_server, sftp_port=self.scp_port,
                                                  username=self.scp_username, password=self.scp_password,
                                                  remote_path=self.bz_image_path)
        logger.info('=== Extract bz_filename from {}: {}'.format(self.bz_image_path,
                                                                 full_bz_filename))
        full_usb_filename = self.scp_file_to_local(local_dir=DUT_DIR_TMP,
                                                   sftp_ip=self.scp_server, sftp_port=self.scp_port,
                                                   username=self.scp_username, password=self.scp_password,
                                                   remote_path=self.usb_image_path)
        logger.info('=== Extract usb_filename from {}: '
                    '{}'.format(self.usb_image_path,
                                full_usb_filename))
        cmd = 'mv {}/{} {}'.format(DUT_DIR_TMP,
                                   full_bz_filename,
                                   DUT_DIR_BOOT)
        self.execute_only(cmd=cmd)

        cmd = 'mv {}/{} {}'.format(DUT_DIR_TMP,
                                   full_usb_filename,
                                   DUT_DIR_BOOT)
        self.execute_only(cmd=cmd)

        # Generate lilo file
        cmd = "cat {} | perl -pe 's/default=[^\n]+/default=INSTALL/g'" \
              " > {}".format(DUT_LILO_TMPL_PATH, DUT_LILO_TMP_PATH)
        self.execute_only(cmd=cmd)

        lilo_append_lines = """
image=/boot/{}
      label=INSTALL
      read-only
      initrd=/boot/{}
      append="root=/dev/ram0 rw PRESERVE_DATA=Yes choicedevice=eth0 INTEGCONF={}/{}"
""".format(full_bz_filename, full_usb_filename, self.http_server, self.config_file_url)
        cmd = '''echo '{}' >> {}'''.format(lilo_append_lines,
                                           DUT_LILO_TMP_PATH)
        self.execute_only(cmd=cmd)

        self.execute('cat {}'.format(DUT_LILO_TMP_PATH))
        self.execute('mv {} {}'.format(DUT_LILO_TMP_PATH, DUT_LILO_PATH))
        logger.info('=== Generated lilo file [{}]\n{}\n'.format(DUT_LILO_PATH,
                                                                lilo_append_lines))
        logger.info('=== run_lilo_and_depmod')
        cmd = '/bin/bash -c "source /etc/lilo.d/functions.lilo; ' \
              'run_lilo_and_depmod"'
        self.execute_only(cmd=cmd, timeout=180)

        self.execute('mount LABEL=SFKEYFOB /usb')
        self.execute('lilo -P ignore -C /etc/lilo.conf.atf')
        self.execute('lilo -q 2>&1')
        logger.info('=== System will reboot!')
        self.spawn_id.sendline('reboot')

    def confirm_device_rebooted(self):
        """Confirm device is rebooting by executing command pwd.

        If device is not reachable, exit. Wait until device is not
        reachable.

        """

        count = 0
        wait_time = 60
        total_count = 20
        while count < total_count:
            try:
                self.execute('pwd')
                count += 1
                logger.info('=== Wait {} seconds for device to reboot ...'.
                            format(wait_time))
                time.sleep(wait_time)
            except:
                logger.info('=== The Appliance at {} is Rebooting.'.format(self.uut_ip))
                break
        if count >= total_count:
            msg = '=== The appliance {} does not appear to have gone down' \
                  ' for reboot. Installer has failed'.format(self.uut_ip)
            logger.info(msg)
            raise RuntimeError(msg)

    def reconnect_and_process_dialog(self, device_reconnect_count, wait_time,
                                     dialog_timeout, start_time, dialog_name, dialog):
        """Reconnect to the device. Process the dialog. Device could reboot
        multiple times. Wait until the device can be accessed and dialog is
        processed.

        :param device_reconnect_count: initial device reconnect count
        :param wait_time: wait time in each loop
        :param dialog_timeout: dialog timeout
        :param start_time: installation start time used to calculate elapsed_time
        :param dialog_name: dialog name used for logging info
        :param dialog: object of Dialog to be processed for device configuration

        """
        from .constants import Series3Constants
        now = datetime.datetime.now()
        elapsed_time = (now - start_time).total_seconds()
        initial_password = self.sm.patterns.default_password
        if elapsed_time < self.installation_timeout:
            logger.info('=== Sleep for {} seconds'.format(wait_time))
        # Once EULA is agreed, device will reboot again

        is_up = True
        while elapsed_time < self.installation_timeout:
            try:
                logger.info('=== Begin: {}'.format(dialog_name))
                time.sleep(wait_time)

                if not self.console:
                    newdev = self.ftd.ssh_vty(ip=Series3Constants.uut_ssh_ip,
                                              port=Series3Constants.uut_ssh_port,
                                              password=initial_password,
                                              timeout=self.ssh_timeout)
                    self.spawn_id = newdev.spawn_id
                    self.sm = newdev.sm
                    device_reconnect_count += 1
                    logger.info('Device reconnected, {} times'.format(device_reconnect_count))
                logger.info('Started dialog process spawn {} timeout {}'.format(self.spawn_id, dialog_timeout))
                dialog.process(self.spawn_id, timeout=dialog_timeout)
                logger.info('=== End: {}'.format(dialog_name))
                return device_reconnect_count
            except Exception as e:
                now = datetime.datetime.now()
                elapsed_time = (now - start_time).total_seconds()
                logger.info('=== Device cannot be accessed, elapsed_time={}, '
                            'wait for {} seconds. Exception {} ...'.format(elapsed_time, wait_time, e))
                is_up = False

        if not is_up:
            raise RuntimeError('=== Waited enough time for device to be up. Seems like it is gone!')

    def poll_device_and_configure(self):
        """Device configuration: agree to EULA, network configuration. Device
        will reboot multiple times.

        :return: None

        """

        start_time = datetime.datetime.now()
        elapsed_time = 0

        if not self.console:
            # Wait for the second reboot
            d0 = Dialog([
                ['Broadcast message from', None, None, False, False],
            ])

            device_reconnect_count0 = self.reconnect_and_process_dialog(
                device_reconnect_count=0,
                wait_time=120,
                dialog_timeout=3600,
                start_time=start_time,
                dialog_name='wait for second reboot',
                dialog=d0)
        else:
            device_reconnect_count0 = 0

        d1 = Dialog([
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
            ['--More--', 'send(q)', None, True, False],
            ["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ",
             'sendline(YES)', None, True, False],
            ['You must change the password for ', None, None, False, False],
        ])

        device_reconnect_count1 = self.reconnect_and_process_dialog(
            device_reconnect_count=device_reconnect_count0,
            wait_time=120,
            dialog_timeout=3600,
            start_time=start_time,
            dialog_name='wait for EULA agreement',
            dialog=d1)

        d2 = Dialog([
            ['Enter new password:', 'sendline({})'.format(self.sm.patterns.login_password),
             None, True, True],
            ['Confirm new password:', 'sendline({})'.format(self.sm.patterns.login_password), None, True,
             False],
            ['Do you want to configure IPv4', 'sendline(y)', None, True, False],
        ])

        if self.uut_ip6 is None:
            d2.append(['Do you want to configure IPv6', 'sendline(n)', None, True, False])
        else:
            d2.append(['Do you want to configure IPv6', 'sendline(y)', None, True, False])

        d2.append(['Configure IPv4 via DHCP or manually', 'sendline(manual)', None, True, False])
        d2.append(['Enter an IPv4 address for the management interface',
                   'sendline({})'.format(self.uut_ip), None, True, False])
        d2.append(['Enter an IPv4 netmask for the management interface',
                   'sendline({})'.format(self.uut_netmask), None, True, False])
        d2.append(['Enter the IPv4 default gateway for the management interface',
                   'sendline({})'.format(self.uut_gateway), None, True, False])

        if self.uut_ip6 is not None:
            d2.append(['Configure IPv6 via DHCP, router, or manually', 'sendline(manual)', None, True, False])
            d2.append(['Enter the IPv6 address for the management interface',
                       'sendline({})'.format(self.uut_ip6), None, True, False])
            d2.append(['Enter the IPv6 address prefix for the management interface',
                       'sendline({})'.format(self.uut_prefix), None, True, False])
            d2.append(['Enter the IPv6 gateway for the management interface',
                       'sendline({})'.format(self.uut_gateway6), None, True, False])

        d2.append(['Enter a fully qualified hostname for this system ',
                   'sendline({})'.format(self.hostname), None, True, False])
        d2.append(['Enter a comma-separated list of DNS servers or',
                   'sendline({})'.format(self.dns_server), None, True, False])
        d2.append(['Enter a comma-separated list of search domains or',
                   'sendline({})'.format(self.search_domains), None, True, False])
        d2.append(['Allow LCD Panel to configure network settings', 'sendline(n)', None, True, False])
        d2.append(['Choose Detection Mode', 'sendline({})'.format(self.detection_mode), None, True, False])
        d2.append(['> ', 'sendline()', None, False, False])

        # must reconnect here. cli_firstboot is started before everything is ready in the ssh case.
        self.reconnect_and_process_dialog(
            device_reconnect_count=device_reconnect_count1,
            wait_time=120,
            dialog_timeout=600,
            start_time=start_time,
            dialog_name='wait for new password',
            dialog=d2)

        logger.info('fully installed.')

    def validate_version(self):
        """Checks if the installed version matches the version in package
        url.

        :return: None

        """

        # check version
        response = self.execute('show version', 30)
        build = re.findall('Build\s(\d+)', response)[0]
        version = re.findall(r'(Version\s){1}([0-9.]+\d)', str(response))[0][1]
        if build in self.version_build and version in self.version_build:
            logger.info('>>>>>> show version result:\n{}\nmatches '
                        'ftd package image: {}'.format(response,
                                                       self.version_build))
            logger.info('Installed ftd version validated')
        else:
            logger.error('Exception: not the same version and build')
            raise RuntimeError('>>>>>> show version result:\n{}\ndoes not match '
                               'ftd package image: {}'.format(response,
                                                              self.version_build))

    def configure_manager(self):
        """Configure manager to be used for registration

        :return: None

        """

        # check version
        if self.manager_nat_id is None:
            response = self.execute('configure manager add {} {}'.format(self.manager,
                                                                         self.manager_key),
                                    120)
        else:
            response = self.execute('configure manager add {} {} {}'.format(self.manager,
                                                                            self.manager_key,
                                                                            self.manager_nat_id),
                                    120)
        matched_obj = re.search('Manager successfully configured.', response)
        if matched_obj is not None:
            logger.info('Configured FTD manager.')
        else:
            logger.error('Exception: configuration of manager failed')
            raise RuntimeError('>>>>>> configure manager result:\n{}\ndoes not match'.format(response))

    def series3_baseline(self, ftd, http_server, scp_server, scp_port,
                         scp_username, scp_password, scp_hostname,
                         version_build, iso_image_path,
                         uut_ip, uut_netmask, uut_gateway, dns_server,
                         hostname='firepower',
                         search_domains='cisco.com',
                         detection_mode='Inline',
                         http_dir_root='/var/www',
                         http_dir_tmp='iso/tmp_dev',
                         scp_dir_root='/nfs/netboot/ims/Release',
                         bz_image='bzImage*',
                         usb_image='usb-ramdisk*',
                         arch='x86_64',
                         uut_ip6=None, uut_prefix=None, uut_gateway6=None,
                         console=False,
                         manager=None, manager_key=None, manager_nat_id=None,
                         timeout=None, http_port='80'):
        """Install FTD iso image on Series 3 Sensor device via http.

        :param ftd: object of Series3 for reconnecting the device
        :param http_server: HTTP Server IP Address, e.g. '10.89.23.80'
                            must use the same http server as the scp server
        :param http_port: HTTP server port, e.g. '80'
        :param scp_server: SCP Server IP Address, e.g. '10.89.23.80'
        :param scp_port: SCP Server Port, e.g. '22'
        :param scp_username: SCP Username
        :param scp_password: SCP Password
        :param scp_hostname: SCP Hostname
        :param version_build: e.g. '6.2.1-1366'
        :param iso_image_path: FTD image path on HTTP server
                                e.g. '/netboot/ims/Development/6.2.1-1366/iso/Sourcefire_3D_Device_S3-6.2.1-1366-Restore.iso'
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param http_dir_root: absolute path of the http root
                                e.g. '/var/www'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. 'iso/tmp_dev'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/nfs/netboot/ims/Release' or '/nfs/netboot/ims/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param console: set to True if the device is accessed through its console;
                defaulted to False
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :return: None

        """
        publish_kick_metric('device.series3.baseline', 1)
        # set baseline timeout
        self.set_installation_timeouts(timeout)
        # Define attributes and variables
        self.ftd = ftd
        self.http_server = http_server
        self.http_port = http_port
        self.scp_server = scp_server
        self.scp_port = scp_port
        self.scp_username = scp_username
        self.scp_password = scp_password
        self.scp_hostname = scp_hostname
        self.version_build = version_build
        self.iso_image_path = iso_image_path
        self.iso_file = iso_image_path.split('/')[-1]
        self.uut_ip = uut_ip
        self.uut_netmask = uut_netmask
        self.uut_gateway = uut_gateway
        self.uut_ip6 = uut_ip6
        self.uut_prefix = uut_prefix
        self.uut_gateway6 = uut_gateway6
        self.dns_server = dns_server
        self.hostname = hostname
        self.search_domains = search_domains
        self.detection_mode = detection_mode
        self.http_dir_root = http_dir_root
        self.http_dir_tmp = http_dir_tmp
        self.http_dir_for_build = '{}/{}/{}'.format(self.http_dir_root,
                                                    self.http_dir_tmp, self.version_build)
        self.config_file = '{}.config'.format(self.iso_file.strip('.iso'))
        self.config_path = '{}/{}'.format(self.http_dir_for_build, self.config_file)
        self.config_file_url = '{}/{}/{}'.format(self.http_dir_tmp, self.version_build, self.config_file)
        self.scp_dir_root = scp_dir_root
        self.bz_image = bz_image
        self.usb_image = usb_image
        self.arch = arch
        self.bz_image_path = '{}/{}/os/{}/boot/{}'.format(self.scp_dir_root,
                                                          self.version_build,
                                                          self.arch, self.bz_image)
        self.usb_image_path = '{}/{}/os/{}/ramdisks/{}'.format(self.scp_dir_root,
                                                               self.version_build,
                                                               self.arch, self.usb_image)
        self.console = console

        self.manager = manager
        self.manager_key = manager_key
        self.manager_nat_id = manager_nat_id

        logger.info('=== Generate config file on the http server ...')
        self.generate_config_file()

        logger.info('=== Copy bz and usb files to dut ...')
        logger.info('=== Generate lilo file on dut ...')
        self.go_to('sudo_state')
        self.copy_bz_usb_files_setup_lilo_to_install()

        if not self.console:
            logger.info('=== Confirm device is rebooted ...')
            self.confirm_device_rebooted()
        else:
            logger.info('=== Console connection available, wait for login prompt ...')
            d = Dialog([
                ['login:', 'sendline({})'.format(self.sm.patterns.login_username),
                 None, False, False],
            ])
            d.process(self.spawn_id, timeout=self.installation_timeout)
            time.sleep(60)
            self.spawn_id.sendline('{}'.format(self.sm.patterns.default_password))

        logger.info('=== Wait for device to be up and configure device ...')
        self.poll_device_and_configure()

        self.go_to('any')

        if self.manager is not None:
            self.configure_manager()

        logger.info('=== Validate version installed ...')
        self.validate_version()

        logger.info('Installation completed successfully.')

    def baseline_by_branch_and_version(self, site, branch, version,
                                       uut_ip, uut_netmask, uut_gateway,
                                       dns_server, iso_file_type='Restore',
                                       serial_port=False, serverIp='', tftpPrefix='',
                                       scpPrefix='', docs='', pxeUsername='', pxePassword='',
                                       pxeHostname='', scpDirRoot='', httpDirRoot='', httpDirTmp='', **kwargs):
        """Baseline series3 sensor by branch and version using PXE servers.
        Look for needed files on devit-engfs, copy them to the local kick server
        and use them to baseline the device.

        :param site: e.g. 'ful', 'ast', 'bgl'
        :param branch: branch name, e.g. 'Release', 'Feature'
        :param version: software build-version, e,g, 6.2.3-623
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param iso_file_type: 'Autotest' or 'Restore'; defaulted to 'Restore'
        :param serial_port: flag used to specify if baseline is done through
            physical serial port; defaulted to 'False'
        :param \**kwargs:
            :Keyword Arguments, any of below optional parameters:
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param http_dir_root: absolute path of the http root
                                e.g. '/tftpboot/asa/'
        :param http_dir_tmp: the temp dir under http_dir_root for version_build
                            and the config file, e.g. '/cache/temp_config'
        :param scp_dir_root: absolute path of the dir for version_build
                            e.g. '/tftpboot/asa/cache/Release' or '/tftpboot/asa/cache/Development'
        :param bz_image: bz image has to be installed on DUT, which is copied
                        from the scp_server, defaulted to 'bzImage*'
        :param usb_image: usb image has to be installed on DUT, which is copied
                            from the scp_server, defaulted to 'usb-ramdisk*'
        :param arch: arch type, defaulted to 'x86_64'
        :param uut_ip6: Device IPv6 Address
        :param uut_prefix: Device IPv6 Prefix
        :param uut_gateway6: Device IPv6 Gateway
        :param console: set to True if the device is accessed through its console;
                defaulted to False
        :param manager: FMC to be configured for registration
        :param manager_key: Registration key
        :param manager_nat_id: Registration NAT Id
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :return: None

        """

        kwargs['uut_ip'] = uut_ip
        kwargs['uut_netmask'] = uut_netmask
        kwargs['uut_gateway'] = uut_gateway
        kwargs['dns_server'] = dns_server
        arch = kwargs.get('arch', 'x86_64')

        if KICK_EXTERNAL:
            server_ip = serverIp
            tftp_prefix = tftpPrefix
            scp_prefix = scpPrefix
            files = docs
        else:
            server_ip, tftp_prefix, scp_prefix, files = prepare_installation_files(site, 's3', branch, version,
                                                                                   iso=iso_file_type, arch_type=arch)
        if not files[0]:
            raise Exception('Series3 iso file not found on server')

        if serial_port:
            iso_file = 'http://{}/{}/{}'.format(server_ip, tftp_prefix[len('asa'):], files[0])
            kwargs['iso_url'] = iso_file
            kwargs['hostname'] = kwargs.get('hostname', 'firepower')
            self.baseline_using_serial(**kwargs)
        else:
            iso_file_name = '{}/{}'.format(tftp_prefix[len('asa'):], files[0])
            kwargs['http_server'] = server_ip
            kwargs['scp_server'] = server_ip
            kwargs['scp_port'] = '22'
            if KICK_EXTERNAL:
                kwargs['scp_username'] = pxeUsername
                kwargs['scp_password'] = pxePassword
                kwargs['scp_hostname'] = pxeHostname
                kwargs['scp_dir_root'] = scpDirRoot
                kwargs['http_dir_root'] = httpDirRoot
                kwargs['http_dir_tmp'] = httpDirTmp
            else:
                kwargs['scp_username'] = pxe_username
                kwargs['scp_password'] = pxe_password
                kwargs['scp_hostname'] = pxe_hostname
                kwargs['scp_dir_root'] = kwargs.get('scp_dir_root',
                                                    '{}/{}'.format(pxe_dir['scp_dir_root'], branch))
                kwargs['http_dir_root'] = kwargs.get('http_dir_root', pxe_dir['http_dir_root'])
                kwargs['http_dir_tmp'] = kwargs.get('http_dir_tmp', pxe_dir['http_dir_tmp'])

            kwargs['iso_image_path'] = iso_file_name
            kwargs['version_build'] = version
            self.series3_baseline(**kwargs)

    def set_installation_timeouts(self, install_timeout=None, ssh_timeout=None):
        """Set the baseline timeout value for this device.

        :param install_timeout: in seconds, timeout for baseline procedure
        :param ssh_timeout: in seconds, timeout for ssh_vty connection
        :return: None
        """

        self.installation_timeout = install_timeout or MAX_TIME_INSTALL
        logger.info('setting baseline max timeout to {}'.format(self.installation_timeout))
        self.ssh_timeout = ssh_timeout or SSH_DEFAULT_TIMEOUT
        logger.info('setting ssh vty max timeout to {}'.format(self.ssh_timeout))

    def _try_to_goto_prompt(self, prompt, num_of_tries=10, raise_execption=True):
        """ Try to go to given prompt in given number of attempts

        :param prompt: Prompt
        :param num_of_tries: Number of attempts to go to given prompt
        :param raise_execption: Raise exception on failure

        :return: return 0 - success
                        1 - failure

        """

        for i in range(1, num_of_tries+1):
            try:
                self.go_to(prompt)
                break
            except:
                pass

        if i >= num_of_tries:
            msg = "Could not go to %s even after %s tries" % (prompt, num_of_tries)
            logger.error(msg)
            if raise_execption:
                raise RuntimeError(msg)

            return 1

        return 0

    def _move_from_lilo_boot_menu_to_lilo_os(self):
        """Go to LILO boot menu to lilo os """

        for i in range(0, 30):
            # Keep sending control+c till device moved to lilo os state
            self.spawn_id.send("\x03")
            try:
                self.spawn_id.expect("Please press Enter to activate this console", timeout=10)
                self.spawn_id.send("\r")
                self.spawn_id.expect(self.sm.patterns.prompt.lilo_os_prompt, timeout=10)
                break
            except:
                pass

        if i >= 30:
            msg = "Could not go to LILO OS prompt from LILO boot prompt"
            logger.error(msg)
            raise RuntimeError(msg)

    def power_cycle(self, power_bar_server, power_bar_port, power_bar_user='admn', power_bar_pwd='admn'):
        """Power cycle the device using power_bar

        :param power_bar_server: comma-separated string of IP addresses of the PDU's
        :param power_bar_port: comma-separated string of power port on the PDU's
        :param power_bar_user: comma-separated string of usernames for power bar servers
        :param power_bar_pwd:  comma-separated string of passwords for power bar servers
        :return:
        """
        if power_bar_server == '' or power_bar_port == '':
            raise AssertionError("Cannot power cycle. powerbar server or powerbar port info missing")

        power_cycle_all_ports(power_bar_server, power_bar_port, power_bar_user, power_bar_pwd)

    def baseline_using_serial(self, iso_url, uut_ip, uut_netmask, uut_gateway, hostname,
                              dns_server='1.1.1.1', search_domains="example.net",
                              detection_mode="Inline", mgmt_intf="eth0", timeout=None, power_cycle_flag=False,
                              pdu_ip='', pdu_port='', pdu_user='admn', pdu_pwd='admn', **kwargs):
        """ Baseline S3 devices through Serial console.

        :param iso_url: HTTP URL for the image.
                    e.g., http://10.83.65.25/cache/Release/6.2.2-81/iso/Sourcefire_3D_Device_S3-6.2.2-81-Restore.iso
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param dns_server: DNS server
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param mgmt_intf: Management Interface name (default: eth0)
        :param timeout: in seconds; time to wait for device to boot with the new image;
                        if not provided, default baseline time is 7200s
        :param power_cycle_flag: True power cylce before baseline, False otherwise
        :param pdu_ip: string of IP addresses of the PDU's
        :param pdu_port: string of power port on the PDU's
        :param pdu_user: usernames for power bar servers
        :param pdu_pwd: passwords for power bar servers
        :return: None

        """
        # set baseline timeout
        self.set_installation_timeouts(timeout)
        if power_cycle_flag:
            self.power_cycle(pdu_ip, pdu_port, pdu_user, pdu_pwd)
            d0 = Dialog([
                ['boot:', None, None, False, False],
            ])
            print(d0.process(self.spawn_id, timeout=600))

            self.spawn_id.send("\t")
            self.spawn_id.expect('boot:')
            self.spawn_id.send("\t")
            self.spawn_id.expect('boot:')

            self.spawn_id.send("Restore_Serial\r")
            self.spawn_id.expect(self.sm.patterns.prompt.lilo_boot_menu_prompt, timeout=120)

            self._move_from_lilo_boot_menu_to_lilo_os()
        else:
            self._try_to_goto_prompt(prompt='any')

            # TODO: Handle eula_prompt and firstboot_state if needed
            # TODO: Add mechanism to handle liloboot_state during login.
            #       It is tough to catch during login as newline or carriage return cause the system to boot.

            # Raise exception if device is not in below states
            baseline_valid_states = ["expert_state", "sudo_state", "prelogin_state", "lilos_state",
                                     "lilobootmenu_state", "fireos_state"]
            if self.sm.current_state not in baseline_valid_states:
                msg = "Baseline Serial function needs one of the valid state from %s" % baseline_valid_states
                logger.error(msg)
                raise RuntimeError(msg)

            # Go to fireos state if serial console is in having active FTD
            if self.sm.current_state in ["expert_state", "sudo_state", "prelogin_state"]:
                self._try_to_goto_prompt(prompt='fireos_state')

            # ---------------------------------------------
            # Bring the system to liloos_state if it is in fireos state
            # ---------------------------------------------
            if self.sm.current_state in ["lilobootmenu_state"]:
                self._move_from_lilo_boot_menu_to_lilo_os()

            elif self.sm.current_state in ["fireos_state"]:
                # term_conn.execute("system reboot", prompt="'YES' or 'NO'")
                # term_conn.spawn_id.sendline("YES")
                self.spawn_id.sendline("system reboot")

                d0 = Dialog([
                    ["This command will reboot the system.  Continue", None, None, True, False],
                    ["'YES' or 'NO':", "sendline(YES)", None, True, False],
                    ['reboot: machine restart', None, None, True, False],
                    ['boot:', None, None, False, False],
                ])

                print(d0.process(self.spawn_id, timeout=600))

                self.spawn_id.send("\t")
                self.spawn_id.expect('boot:')
                self.spawn_id.send("\t")
                self.spawn_id.expect('boot:')

                self.spawn_id.send("Restore_Serial\r")
                self.spawn_id.expect(self.sm.patterns.prompt.lilo_boot_menu_prompt, timeout=120)

                self._move_from_lilo_boot_menu_to_lilo_os()

        # Here means device is in LILO OS State
        self.sm.update_cur_state('lilos_state')

        # Try install seq with retry for 5 times
        for i in range(1, 5):
            ret = self._install_from_iso(iso_url=iso_url, uut_ip=uut_ip, uut_netmask=uut_netmask,
                                         uut_gateway=uut_gateway, mgmt_intf=mgmt_intf)

            if ret == 0:
                break

            # If the device moved back to lilo boot menu. Try to go to lilo os and try install again
            if ret == 1:
                self._move_from_lilo_boot_menu_to_lilo_os()

        # Raise error in all the attempts failed
        if i >= 5:
            msg = "Failed to install S3 using serial for %s times" % i
            logger.error(msg)
            raise RuntimeError(msg)

        time.sleep(60)

        # Perform Initialization
        for i in range(1, 4):
            try:
                self._initialize_dev_after_baseline(uut_ip=uut_ip, uut_netmask=uut_netmask,
                                                    uut_gateway=uut_gateway, dns_server=dns_server,
                                                    hostname=hostname, search_domains=search_domains,
                                                    detection_mode=detection_mode, dhcp=False)
                break
            except:
                logger.error("Failed to perform system initialization")
                raise RuntimeError("Failed to perform system initialization")

    def _install_from_iso(self, iso_url, uut_ip, uut_netmask, uut_gateway, mgmt_intf="eth0", usb_loc="/dev/sdb1"):
        """ Install image from given ISO

        :param iso_url: ISO URL
        :param uut_ip:  Device IP
        :param uut_netmask: Device netmask
        :param uut_gateway: Devcie Gateway
        :param mgmt_intf: Management Interface name (default: eth0)
        :param usb_loc: USB location in devcie (default: /dev/sdb1)
        :return: None

        """

        p = self.sm.patterns.prompt.lilo_os_prompt

        self.spawn_id.sendline("")
        self.spawn_id.expect(self.sm.patterns.prompt.lilo_os_prompt)

        self.execute("mount %s /mnt/usb" % usb_loc)
        self.execute("/sbin/ifconfig %s %s netmask %s" % (mgmt_intf, uut_ip, uut_netmask))
        self.execute("route add default gw %s" % uut_gateway)

        # Get ISO name, path and server ip from URL path
        url_data = urllib.parse.urlparse(iso_url)
        install_srv = url_data.netloc
        iso_name = os.path.basename(url_data.path)
        iso_path = os.path.dirname(url_data.path)[1:]

        self.execute("""echo "SRV=%s
        dirpath=%s
        TRANS=httpsvr
        choicedevice=%s
        IPCONF=dhcp
        IPFAM=ipv4
        choiceiso=%s
        " > /mnt/usb/configs/default_config.conf""" % (install_srv, iso_path, mgmt_intf, iso_name))

        self.execute("cat /mnt/usb/configs/default_config.conf")

        self.execute("umount /cdrom  2> /dev/null")
        self.execute("rm -f /mnt/tmpfs/*")
        self.execute("umount /mnt/tmpfs 2> /dev/null")
        self.execute("umount /mnt/hd/ 2> /dev/null")
        self.execute("umount /mnt/usb 2> /dev/null")

        self.execute("/etc/rc.d/install.sh --partition-only", timeout=240)

        # Download image into /mnt/var and initiate installation
        self.execute("cd /mnt/var/")

        for i in range(1, 11):
            try:
                d1 = Dialog([
                    ['The file is already fully retrieved; nothing to do.', None, None, False, False],
                    ["'%s' saved" % iso_name, None, None, False, False],
                ])
                self.spawn_id.sendline("wget -c %s" % iso_url)
                d1.process(self.spawn_id, timeout=300)
                break
            except:
                pass

        if i >= 10:
            msg = "Failed to down load %s in 10 attempts with timeout 5 mins" % "wget -c %s" % iso_url
            logger.error(msg)
            raise RuntimeError(msg)

        self.execute("mount -t iso9660 -o loop /mnt/var/%s /cdrom" % iso_name, timeout=90)

        # Install steps
        self.execute("umount /mnt/usb &> /dev/null")

        self.spawn_id.sendline("/cdrom/install/menu.sh")

        d0 = Dialog([
            ["The system will restart after you press enter.", 'sendline()', None, True, False],
            ['Restore the system.*yes/no.*:', 'sendline(yes)', None, True, False],
            ['Delete license and network settings.*yes/no.*:', 'sendline(no)', None, True, False],
            ['Are you sure.*yes/no.*:', 'sendline(yes)', None, True, False],
            [p, None, None, False, False],
            [self.sm.patterns.prompt.lilo_boot_menu_prompt, None, None, False, False],
        ])
        res = d0.process(self.spawn_id, timeout=600)
        print(dir(res))

        if self.sm.patterns.prompt.lilo_boot_menu_prompt in res.match_output:
            logger.info("Device moved back to lilo boot menu")
            return 1

        if "root@(none)" in res.match_output:
            self.execute("umount /cdrom")
            self.execute("losetup -d /dev/loop0")
            self.execute("rm -rf /mnt/var/*.iso")

            # Reboot system to complete install process
            self.spawn_id.sendline("reboot")
            d0 = Dialog([
                ["firepower login:", None, None, False, False],
                [self.sm.patterns.prompt.prelogin_prompt, None, None, False, False],
                [self.sm.patterns.prompt.lilo_boot_menu_prompt, None, None, False, False],
            ])
            res = d0.process(self.spawn_id, timeout=self.installation_timeout)

            if self.sm.patterns.prompt.lilo_boot_menu_prompt in res.match_output:
                logger.info("Device moved back to lilo boot menu")
                return 1
            elif 'login:' in res.match_output:
                print("Install is done and got firepower login: prompt")
                logger.info("Install is done and got firepower login: prompt")
                self.sm.update_cur_state('prelogin_state')
                return 0

        logger.error("Unknown state. It should be never be here")

        return 2

    def _initialize_dev_after_baseline(self, dhcp=True, uut_ip=None, uut_netmask=None, uut_gateway=None,
                                       dns_server='1.1.1.1', hostname=None, search_domains="example.net",
                                       detection_mode="Inline"):
        """ Perform System Initialization after installation.

        :param dhcp: Is IP from DHCP. True- from dhcp, False - manual
        :param uut_ip: Device IP Address
        :param uut_netmask: Device Netmask
        :param uut_gateway: Device Gateway
        :param dns_server: DNS server
        :param hostname: hostname to be set, e.g. 'BATIT-3D8130-1-AST'
        :param search_domains: search domains delimited by comma,
                                defaulted to 'cisco.com'
        :param detection_mode: defaulted to 'Inline',
                                can be: 'Passive','Inline','Access Control','Network Discovery'
        :param mgmt_intf: Management Interface name (default: eth0)
        :return: None

        """

        # TODO: Handling IPv6 address configuration

        # Perform Initialization
        self.spawn_id.sendline("")

        d2 = Dialog([
            ['Press <ENTER> to display the EULA: ', 'sendline()', None, True, False],
        ])
        d2.append(['--More--', 'send(q)', None, True, False])
        d2.append(["Please.*enter.*YES.*AGREE.*EULA:", 'sendline(YES)', None, True, False])
        d2.append(["Please enter 'YES' or press <ENTER> to AGREE to the EULA: ", 'sendline(YES)', None, True, False])
        d2.append(['Enter new password:', 'sendline({})'.format(self.sm.patterns.login_password), None, True, True])
        d2.append(['Confirm new password:', 'sendline({})'.format(self.sm.patterns.login_password), None, True, False])

        d2.append(['Do you want to configure IPv4', 'sendline(y)', None, True, False])
        d2.append(['Do you want to configure IPv6', 'sendline(n)', None, True, False])

        if dhcp:
            d2.append(['Configure IPv4 via DHCP or manually', 'sendline(dhcp)', None, True, False])
        else:
            d2.append(['Configure IPv4 via DHCP or manually', 'sendline(manual)', None, True, False])

            d2.append(['Enter an IPv4 address for the management interface', 'sendline({})'.format(uut_ip),
                       None, True, False])
            d2.append(['Enter an IPv4 netmask for the management interface', 'sendline({})'.format(uut_netmask),
                       None, True, False])
            d2.append(['Enter the IPv4 default gateway for the management interface', 'sendline({})'.format(uut_gateway),
                       None, True, False])
            d2.append(['Enter a fully qualified hostname for this system ', 'sendline({})'.format(hostname),
                       None, True, False])
            d2.append(['Enter a comma-separated list of DNS servers or', 'sendline({})'.format(dns_server),
                       None, True, False])
            d2.append(['Enter a comma-separated list of search domains or', 'sendline({})'.format(search_domains),
                       None, True, False])

        d2.append(['Allow LCD Panel to configure network settings', 'sendline(n)', None, True, False])
        d2.append(['Choose Detection Mode', 'sendline({})'.format(detection_mode), None, False, False])
        d2.append(['firepower login:', "sendline({})".format(self.sm.patterns.username), None, True, False])
        d2.append(['{} login:'.format(hostname), "sendline({})".format(self.sm.patterns.username), None, True, False])
        d2.append(['[P|p]assword:', "sendline({})".format(self.sm.patterns.default_password), None, True, False])

        res = d2.process(self.spawn_id, timeout=720)

        time.sleep(20)
        self.spawn_id.sendline("")
        self.spawn_id.expect(">", timeout=300)

    def replace_asa_image(self, source_location, pwd, timeout=300):
        """ Not implemented for Series3Line class"""

        raise NotImplementedError




from unicon.eal.dialogs import Statement


def login_handler(spawn, patterns):
    spawn.sendline(patterns.login_username)
    spawn.expect('Password: ')
    spawn.sendline(patterns.login_password)
    spawn.expect('Cisco Firepower')
    spawn.sendline()


class SspStatements:
    def __init__(self, patterns):
        self.login_password = Statement(pattern=patterns.prompt.prelogin_prompt, action=login_handler,
                                        args={'patterns': patterns}, loop_continue=True, continue_timer=True)
